"""JSON apply_filter tests."""
from __future__ import annotations

__all__: list[str] = []

from typing import TYPE_CHECKING, Any

import pytest

from jsonyx import JSONSyntaxError, apply_filter
from jsonyx.test import check_syntax_err

if TYPE_CHECKING:
    _Node = tuple[dict[Any, Any] | list[Any], Any]


@pytest.mark.parametrize(("node", "keep"), [
    # List
    (([], 0), False),
    (([0], 0), True),

    # Dict
    (({}, ""), False),
    (({"": 0}, ""), True),
])  # type: ignore
def test_has_key(node: _Node, keep: bool) -> None:
    """Test has key."""
    expected: list[_Node] = [node] if keep else []
    assert apply_filter(node, "@") == expected


@pytest.mark.parametrize(("node", "keep"), [
    # List
    (([], 0), True),
    (([0], 0), False),

    # Dict
    (({}, ""), True),
    (({"": 0}, ""), False),
])  # type: ignore
def test_has_not_key(node: _Node, keep: bool) -> None:
    """Test has not key."""
    expected: list[_Node] = [node] if keep else []
    assert apply_filter(node, "!@") == expected


@pytest.mark.parametrize(("query", "keep"), [
    # Less than or equal to
    ("@ <= -1", False),
    ("@ <= 0", True),
    ("@ <= 1", True),

    # Less than
    ("@ < -1", False),
    ("@ < 0", False),
    ("@ < 1", True),

    # Equal to
    ("@ == 1", False),
    ("@ == 0", True),

    # Not equal to
    ("@ != 0", False),
    ("@ != 1", True),

    # Greater than or equal to
    ("@ >= 1", False),
    ("@ >= 0", True),
    ("@ >= -1", True),

    # Greater than
    ("@ > 1", False),
    ("@ > 0", False),
    ("@ > -1", True),
])
def test_operator(query: str, keep: bool) -> None:
    """Test operator."""
    expected: list[_Node] = [([0], 0)] if keep else []
    assert apply_filter(([0], 0), query) == expected


@pytest.mark.parametrize("query", [
    # No whitespace
    "@==0",

    # Before operator
    "@ ==0",

    # After operator
    "@== 0",
])
def test_operator_whitespace(query: str) -> None:
    """Test whitespace around operator."""
    assert not apply_filter([], query)


def test_and() -> None:
    """Test and."""
    query: str = "@ != 1 && @ != 2 && @ != 3"
    assert apply_filter(([0], 0), query) == [([0], 0)]


@pytest.mark.parametrize("query", [
    # No whitespace
    "@!=1&&@!=2&&@!=3",

    # Before and
    "@!=1 &&@!=2 &&@!=3",

    # After and
    "@!=1&& @!=2&& @!=3",
])
def test_and_whitespace(query: str) -> None:
    """Test whitespace around and."""
    assert apply_filter(([0], 0), query) == [([0], 0)]


@pytest.mark.parametrize(("query", "msg", "colno", "end_colno"), [
    ("", "Expecting a relative query", 1, -1),
    ("@ == ", "Expecting value", 6, -1),
    ("@ && ", "Expecting a relative query", 6, -1),
    ("!@ == 0", "Unexpected operator", 4, 6),
    ("@ @ @", "Expecting end of file", 2, -1),
])
def test_invalid_query(
    query: str, msg: str, colno: int, end_colno: int,
) -> None:
    """Test invalid query."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        apply_filter([], query)

    check_syntax_err(exc_info, msg, colno, end_colno)


def test_slice() -> None:
    """Test slice."""
    with pytest.raises(TypeError, match="List index must be int"):
        apply_filter(([[]], 0), "@[:]")



"""JSON load_query_value tests."""
from __future__ import annotations

__all__: list[str] = []

from decimal import MAX_EMAX, Decimal

import pytest

from jsonyx import JSONSyntaxError, load_query_value
from jsonyx.allow import NAN_AND_INFINITY
from jsonyx.test import check_syntax_err


@pytest.mark.parametrize(("s", "expected"), [
    ("true", True),
    ("false", False),
    ("null", None),
])
def test_literal_names(s: str, expected: bool | None) -> None:
    """Test literal names."""
    assert load_query_value(s) is expected


@pytest.mark.parametrize("s", ["Infinity", "-Infinity"])
@pytest.mark.parametrize("use_decimal", [True, False])
def test_infinity(s: str, use_decimal: bool) -> None:
    """Test infinity."""
    obj: object = load_query_value(
        s, allow=NAN_AND_INFINITY, use_decimal=use_decimal,
    )
    expected_type: type[Decimal | float] = Decimal if use_decimal else float
    assert isinstance(obj, expected_type)
    assert obj == expected_type(s)


@pytest.mark.parametrize("s", ["Infinity", "-Infinity"])
@pytest.mark.parametrize("use_decimal", [True, False])
def test_infinity_not_allowed(s: str, use_decimal: bool) -> None:
    """Test infinity when not allowed."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        load_query_value(s, use_decimal=use_decimal)

    check_syntax_err(exc_info, f"{s} is not allowed", 1, len(s) + 1)


@pytest.mark.parametrize("s", [
    # Sign
    "-1",

    # Integer
    "0", "1", "10", "11",
])
def test_int(s: str) -> None:
    """Test integer."""
    obj: object = load_query_value(s)
    assert isinstance(obj, int)
    assert obj == int(s)


def test_too_big_int(big_num: str) -> None:
    """Test too big integer."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        load_query_value(big_num)

    check_syntax_err(exc_info, "Invalid number", 1, len(big_num) + 1)


@pytest.mark.parametrize("s", [
    # Sign
    "-1.0",

    # Fraction
    "1.0", "1.01", "1.1", "1.11",

    # Fraction with trailing zeros
    "1.00", "1.10",

    # Exponent e
    "1E1",

    # Exponent sign
    "1e-1", "1e+1",

    # Exponent power
    "1e0", "1e1", "1e10", "1e11",

    # Exponent power with leading zeros
    "1e00", "1e01",

    # Parts
    "1.1e1", "-1e1", "-1.1", "-1.1e1",

    # Big exponent
    "1e400", "-1e400",
])
@pytest.mark.parametrize("use_decimal", [True, False])
def test_rational_number(s: str, use_decimal: bool) -> None:
    """Test rational number."""
    obj: object = load_query_value(s, use_decimal=use_decimal)
    expected_type: type[Decimal | float] = Decimal if use_decimal else float
    assert isinstance(obj, expected_type)
    assert obj == expected_type(s)


@pytest.mark.parametrize("s", [f"1e{MAX_EMAX + 1}", f"-1e{MAX_EMAX + 1}"])
def test_too_big_number(s: str) -> None:
    """Test too big JSON number."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        load_query_value(s, use_decimal=True)

    check_syntax_err(exc_info, "Invalid number", 1, len(s) + 1)


@pytest.mark.parametrize("s", ["1\uff10", "0.\uff10", "0e\uff10"])
def test_invalid_number(s: str) -> None:
    """Test invalid number."""
    with pytest.raises(JSONSyntaxError):
        load_query_value(s)


@pytest.mark.parametrize("s", [
    # Empty string
    "",

    # One character
    "$",

    # Multiple characters
    "foo",
])
def test_string(s: str) -> None:
    """Test JSON string."""
    assert load_query_value(f"'{s}'") == s


@pytest.mark.parametrize(("s", "expected"), [
    # Tilde escapes
    ("~'", "'"),
    ("~~", "~"),

    # Multiple characters
    ("foo~~bar", "foo~bar"),
])
def test_string_escapes(s: str, expected: str) -> None:
    """Test string escapes."""
    assert load_query_value(f"'{s}'") == expected


@pytest.mark.parametrize(("s", "msg", "colno", "end_colno"), [
    ("'foo", "Unterminated string", 1, 5),
    ("'~", "Expecting escaped character", 3, -1),
    ("'~a'", "Invalid tilde escape", 2, 4),
])
def test_invalid_string(s: str, msg: str, colno: int, end_colno: int) -> None:
    """Test invalid JSON string."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        load_query_value(s)

    check_syntax_err(exc_info, msg, colno, end_colno)


@pytest.mark.parametrize("s", ["", "NaN"])
def test_expecting_value(s: str) -> None:
    """Test expecting JSON value."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        load_query_value(s)

    check_syntax_err(exc_info, "Expecting value")


def test_end_of_file() -> None:
    """Test end of file."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        load_query_value("1 2 3")

    check_syntax_err(exc_info, "Expecting end of file", 2)



"""Allow minimal and unambiguous JSON deviations for developer convenience.

See :doc:`the specification </jsonyx-spec>` for more information.

.. tip:: Flags can also be combined (e.g. ``COMMENTS | UNQUOTED_KEYS``) or
    excluded (e.g. ``EVERYTHING - NON_STR_KEYS``).
"""
from __future__ import annotations

__all__: list[str] = [
    "COMMENTS",
    "EVERYTHING",
    "MISSING_COMMAS",
    "NAN_AND_INFINITY",
    "NON_STR_KEYS",
    "NOTHING",
    "SURROGATES",
    "TRAILING_COMMA",
    "UNQUOTED_KEYS",
]

NOTHING: frozenset[str] = frozenset()
"""Raise an error for all deviations from :rfc:`8259`."""

COMMENTS: frozenset[str] = frozenset({"comments"})
"""Allow block and line comments.

>>> import jsonyx as json
>>> import jsonyx.allow
>>> json.loads("0 /* Block */ // and line comment", allow=jsonyx.allow.COMMENTS)
0
"""

MISSING_COMMAS: frozenset[str] = frozenset({"missing_commas"})
"""Allow separating items with whitespace.

>>> import jsonyx as json
>>> import jsonyx.allow
>>> json.loads("[1 2 3]", allow=jsonyx.allow.MISSING_COMMAS)
[1, 2, 3]
"""

NAN_AND_INFINITY: frozenset[str] = frozenset({"nan_and_infinity"})
"""Allow ``NaN``, ``Infinity`` and ``-Infinity``.

>>> import jsonyx as json
>>> import jsonyx.allow
>>> json.loads("[NaN, Infinity, -Infinity]", allow=jsonyx.allow.NAN_AND_INFINITY)
[nan, inf, -inf]
>>> from math import inf, nan
>>> json.dump([nan, inf, -inf], allow=jsonyx.allow.NAN_AND_INFINITY)
[NaN, Infinity, -Infinity]

.. note:: ``Decimal("sNan")`` can't be (de)serialised this way.
"""

NON_STR_KEYS: frozenset[str] = frozenset({"non_str_keys"})
"""Allow non-string keys in ``"object"`` types.

.. versionadded:: 2.0

>>> import jsonyx as json
>>> jsonyx.allow
>>> json.dump({1: 0}, allow=jsonyx.allow.NON_STR_KEYS)
{"1": 0}

.. warning:: This can result in duplicate keys if a non-string key is
    serialised to the same string as an existing key.
"""

SURROGATES: frozenset[str] = frozenset({"surrogates"})
r"""Allow unpaired surrogates in strings.

>>> import jsonyx as json
>>> import jsonyx.allow
>>> json.loads('"\ud800"', allow=jsonyx.allow.SURROGATES)
'\ud800'
>>> json.dump("\ud800", allow=jsonyx.allow.SURROGATES, ensure_ascii=True)
"\ud800"

.. tip:: If you're not using ``read()`` or ``write()``, you still need to
    use ``ensure_ascii=True`` or set the unicode error handler to
    ``"surrogatepass"``.
"""

TRAILING_COMMA: frozenset[str] = frozenset({"trailing_comma"})
"""Allow a trailing comma at the end of arrays and objects.

>>> import jsonyx as json
>>> import jsonyx.allow
>>> json.loads("[0,]", allow=jsonyx.allow.TRAILING_COMMA)
[0]
"""

UNQUOTED_KEYS: frozenset[str] = frozenset({"unquoted_keys"})
"""Allow unquoted keys in objects which are :ref:`identifiers <identifiers>`.

.. versionadded:: 2.0

>>> import jsonyx as json
>>> import jsonyx.allow
>>> json.loads('{key: "value"}', allow=jsonyx.allow.UNQUOTED_KEYS)
{'key': 'value'}
"""

EVERYTHING: frozenset[str] = (
    COMMENTS | MISSING_COMMAS | NAN_AND_INFINITY | NON_STR_KEYS | SURROGATES
    | TRAILING_COMMA | UNQUOTED_KEYS
)
"""Allow all JSON deviations.

This is equivalent to ``COMMENTS | MISSING_COMMAS | NAN_AND_INFINITY
| NON_STR_KEYS | SURROGATES | TRAILING_COMMA | UNQUOTED_KEYS``.
"""



"""JSON write tests."""
from __future__ import annotations

__all__: list[str] = []

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING

import pytest

from jsonyx.allow import SURROGATES

if TYPE_CHECKING:
    from types import ModuleType


def test_value(json: ModuleType) -> None:
    """Test JSON value."""
    with TemporaryDirectory() as tmpdir:
        filename: Path = Path(tmpdir) / "file.json"
        json.write(0, filename, end="")
        assert filename.read_text("utf-8") == "0"


@pytest.mark.parametrize("s", ["\ud800", "\ud800$", "\udf48"])  # noqa: PT014
def test_surrogates(json: ModuleType, s: str) -> None:
    """Test surrogates."""
    with TemporaryDirectory() as tmpdir:
        filename: Path = Path(tmpdir) / "file.json"
        json.write(s, filename, allow=SURROGATES, end="")
        assert filename.read_text("utf-8", "surrogatepass") == f'"{s}"'


@pytest.mark.parametrize("s", ["\ud800", "\ud800$", "\udf48"])  # noqa: PT014
def test_surrogates_if_not_allowed(json: ModuleType, s: str) -> None:
    """Test surrogates."""
    with TemporaryDirectory() as tmpdir, \
            pytest.raises(json.TruncatedSyntaxError):
        json.write(s, Path(tmpdir) / "file.json")



"""JSON dump tests."""
from __future__ import annotations

__all__: list[str] = []

from contextlib import redirect_stdout
from io import StringIO
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import ModuleType


def test_file(json: ModuleType) -> None:
    """Test write to file."""
    io: StringIO = StringIO()
    json.dump(0, io, end="")
    assert io.getvalue() == "0"


def test_stdout(json: ModuleType) -> None:
    """Test write to stdout."""
    with redirect_stdout(StringIO()) as io:
        json.dump(0, end="")
        assert io.getvalue() == "0"



"""Customizable JSON library for Python."""
from __future__ import annotations

__all__: list[str] = [
    "Decoder",
    "Encoder",
    "JSONSyntaxError",
    "Manipulator",
    "TruncatedSyntaxError",
    "apply_filter",
    "apply_patch",
    "detect_encoding",
    "dump",
    "dumps",
    "format_syntax_error",
    "load",
    "load_query_value",
    "loads",
    "make_patch",
    "read",
    "select_nodes",
    "write",
]
__version__: str = "2.3.0"

from typing import TYPE_CHECKING, Any, Protocol, TypeVar

from jsonyx._decoder import (
    Decoder, JSONSyntaxError, TruncatedSyntaxError, detect_encoding,
)
from jsonyx._differ import make_patch
from jsonyx._encoder import Encoder
from jsonyx._manipulator import Manipulator
from jsonyx.allow import NOTHING

if TYPE_CHECKING:
    from collections.abc import Callable, Container
    from os import PathLike

    _T_co = TypeVar("_T_co", covariant=True)
    _T_contra = TypeVar("_T_contra", contravariant=True)

    # pylint: disable-next=R0903
    class _SupportsRead(Protocol[_T_co]):
        def read(self, length: int = ..., /) -> _T_co:  # type: ignore
            """Read string."""

    # pylint: disable-next=R0903
    class _SupportsWrite(Protocol[_T_contra]):
        def write(self, s: _T_contra, /) -> object:
            """Write string."""

    _Node = tuple[dict[Any, Any] | list[Any], Any]
    _Operation = dict[str, Any]
    _StrPath = PathLike[str] | str
    _Hook = Callable[[Any], Any]


def format_syntax_error(exc: TruncatedSyntaxError) -> list[str]:
    """Format a truncated syntax error.

    :param exc: a truncated syntax error
    :return: a list of strings, each ending in a newline

    Example:
        >>> import jsonyx as json
        >>> try:
        ...     json.loads("[,]")
        ... except json.JSONSyntaxError as exc:
        ...     print("Traceback (most recent call last):")
        ...     print(end="".join(json.format_syntax_error(exc)))
        ...
        Traceback (most recent call last):
          File "<string>", line 1, column 2
            [,]
             ^
        jsonyx.JSONSyntaxError: Expecting value

    .. note:: Don't use :func:`traceback.format_exception_only`, it displays
        less information.

    """
    if exc.end_lineno == exc.lineno:
        line_range: str = f"{exc.lineno:d}"
    else:
        line_range = f"{exc.lineno:d}-{exc.end_lineno:d}"

    if exc.end_colno == exc.colno:
        column_range: str = f"{exc.colno:d}"
    else:
        column_range = f"{exc.colno:d}-{exc.end_colno:d}"

    indent: str = " " * (exc.offset - 1)  # type: ignore
    if exc.end_lineno == exc.lineno:
        selection: str = "^" * (exc.end_offset - exc.offset)  # type: ignore
    else:
        selection = "^" * (len(exc.text) - exc.offset + 1)  # type: ignore

    return [
        f'  File "{exc.filename}", line {line_range}, column {column_range}\n',
        f"    {exc.text}\n",
        f"    {indent}{selection}\n",
        f"{exc.__module__}.{type(exc).__qualname__}: {exc.msg}\n",
    ]


def read(
    filename: _StrPath,
    *,
    allow: Container[str] = NOTHING,
    cache_keys: bool = False,
    hooks: dict[str, _Hook] | None = None,
) -> Any:
    """Deserialize a JSON file to a Python object.

    .. versionchanged:: 2.0 Replaced ``use_decimal`` with ``hooks``.

    .. versionchanged:: 2.2

        - Added ``cache_keys``.
        - Disabled caching keys by default.

    :param filename: the path to the JSON file
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param cache_keys: re-use the keys of objects
    :param hooks: the :ref:`hooks <decoding_hooks>` used for transforming data
    :raises OSError: if the file can't be opened
    :raises TruncatedSyntaxError: when failing to decode the file
    :return: a Python object.

    Example:
        >>> import jsonyx as json
        >>> from pathlib import Path
        >>> from tempfile import TemporaryDirectory
        >>> with TemporaryDirectory() as tmpdir:
        ...     filename = Path(tmpdir) / "file.json"
        ...     _ = filename.write_text('["filesystem API"]', "utf-8")
        ...     json.read(filename)
        ...
        ['filesystem API']

    """
    return Decoder(
        allow=allow, cache_keys=cache_keys, hooks=hooks,
    ).read(filename)


def load(
    fp: _SupportsRead[bytes | str],
    *,
    allow: Container[str] = NOTHING,
    cache_keys: bool = False,
    hooks: dict[str, _Hook] | None = None,
    root: _StrPath = ".",
) -> Any:
    """Deserialize an open JSON file to a Python object.

    .. versionchanged:: 2.0 Replaced ``use_decimal`` with ``hooks``.

    .. versionchanged:: 2.2

        - Added ``cache_keys``.
        - Disabled caching keys by default.

    :param fp: an open JSON file
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param cache_keys: re-use the keys of objects
    :param hooks: the :ref:`hooks <decoding_hooks>` used for transforming data
    :param root: the path to the archive containing this JSON file
    :raises TruncatedSyntaxError: when failing to decode the file
    :return: a Python object

    Example:
        >>> import jsonyx as json
        >>> from io import StringIO
        >>> io = StringIO('["streaming API"]')
        >>> json.load(io)
        ['streaming API']

    .. tip:: Specify ``root`` to display the zip filename in error messages.

    """
    return Decoder(
        allow=allow, cache_keys=cache_keys, hooks=hooks,
    ).load(fp, root=root)


def loads(
    s: bytearray | bytes | str,
    *,
    allow: Container[str] = NOTHING,
    cache_keys: bool = False,
    filename: _StrPath = "<string>",
    hooks: dict[str, _Hook] | None = None,
) -> Any:
    r"""Deserialize a JSON string to a Python object.

    .. versionchanged:: 2.0 Replaced ``use_decimal`` with ``hooks``.

    .. versionchanged:: 2.2

        - Added ``cache_keys``.
        - Disabled caching keys by default.

    :param s: a JSON string
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param cache_keys: re-use the keys of objects
    :param filename: the path to the JSON file
    :param hooks: the :ref:`hooks <decoding_hooks>` used for transforming data
    :raises TruncatedSyntaxError: when failing to decode the string
    :return: a Python object

    Example:
        >>> import jsonyx as json
        >>> json.loads('{"foo": ["bar", null, 1.0, 2]}')
        {'foo': ['bar', None, 1.0, 2]}
        >>> json.loads(r'"\"foo\bar"')
        '"foo\x08ar'

    .. tip:: Specify ``filename`` to display the filename in error messages.

    """
    return Decoder(
        allow=allow, cache_keys=cache_keys, hooks=hooks,
    ).loads(s, filename=filename)


def write(
    obj: object,
    filename: _StrPath,
    encoding: str = "utf-8",
    *,
    allow: Container[str] = NOTHING,
    check_circular: bool = True,
    commas: bool = True,
    end: str = "\n",
    ensure_ascii: bool = False,
    hook: _Hook | None = None,
    indent: int | str | None = None,
    indent_leaves: bool = True,
    max_indent_level: int | None = None,
    quoted_keys: bool = True,
    separators: tuple[str, str] = (", ", ": "),
    skipkeys: bool = False,
    sort_keys: bool = False,
    trailing_comma: bool = False,
    types: dict[str, type | tuple[type, ...]] | None = None,
) -> None:
    r"""Serialize a Python object to a JSON file.

    .. versionchanged:: 2.0

        - Added ``commas``, ``encoding``, ``indent_leaves``,
          ``max_indent_level``, ``quoted_keys`` and ``types``.
        - Made :class:`tuple` serializable by default instead of
          :class:`enum.Enum` and :class:`decimal.Decimal`.
        - Replaced ``item_separator`` and ``key_separator`` with
          ``separators``.

    .. versionchanged:: 2.1 Added ``check_circular``, ``hook`` and
        ``skipkeys``.

    :param obj: a Python object
    :param filename: the path to the JSON file
    :param encoding: the JSON encoding
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param check_circular: check for circular references
    :param commas: separate items by commas when indented
    :param end: the string to append at the end
    :param ensure_ascii: escape non-ASCII characters
    :param hook: the :ref:`hook <encoding_hook>` used for transforming data
    :param indent: the number of spaces or string to indent with
    :param indent_leaves: indent leaf objects and arrays
    :param max_indent_level: the level up to which to indent
    :param quoted_keys: quote keys which are :ref:`identifiers <identifiers>`
    :param separators: the item and key separator
    :param skipkeys: skip non-string keys
    :param sort_keys: sort the keys of objects
    :param trailing_comma: add a trailing comma when indented
    :param types: a dictionary of :ref:`additional types <protocol_types>`
    :raises OSError: if the file can't be opened
    :raises RecursionError: if the object is too deeply nested
    :raises TypeError: for unserializable values
    :raises TruncatedSyntaxError: when failing to encode the file
    :raises ValueError: for invalid values

    Example:
        >>> import jsonyx as json
        >>> from pathlib import Path
        >>> from tempfile import TemporaryDirectory
        >>> with TemporaryDirectory() as tmpdir:
        ...     filename = Path(tmpdir) / "file.json"
        ...     json.write(["filesystem API"], filename)
        ...     filename.read_text("utf-8")
        ...
        '["filesystem API"]\n'

    .. note:: The item separator is automatically stripped when indented.
    .. warning:: Avoid specifying ABCs for ``types``, that is very slow.

    """
    Encoder(
        allow=allow,
        check_circular=check_circular,
        commas=commas,
        end=end,
        ensure_ascii=ensure_ascii,
        hook=hook,
        indent=indent,
        indent_leaves=indent_leaves,
        max_indent_level=max_indent_level,
        quoted_keys=quoted_keys,
        separators=separators,
        skipkeys=skipkeys,
        sort_keys=sort_keys,
        trailing_comma=trailing_comma,
        types=types,
    ).write(obj, filename, encoding)


def dump(
    obj: object,
    fp: _SupportsWrite[str] | None = None,
    *,
    allow: Container[str] = NOTHING,
    check_circular: bool = True,
    commas: bool = True,
    end: str = "\n",
    ensure_ascii: bool = False,
    hook: _Hook | None = None,
    indent: int | str | None = None,
    indent_leaves: bool = True,
    max_indent_level: int | None = None,
    quoted_keys: bool = True,
    separators: tuple[str, str] = (", ", ": "),
    skipkeys: bool = False,
    sort_keys: bool = False,
    trailing_comma: bool = False,
    types: dict[str, type | tuple[type, ...]] | None = None,
) -> None:
    r"""Serialize a Python object to an open JSON file.

    .. versionchanged:: 2.0

        - Added ``commas``, ``indent_leaves``, ``max_indent_level``,
          ``quoted_keys`` and ``types``.
        - Made :class:`tuple` serializable by default instead of
          :class:`enum.Enum` and :class:`decimal.Decimal`.
        - Replaced ``item_separator`` and ``key_separator`` with
          ``separators``.

    .. versionchanged:: 2.1 Added ``check_circular``, ``hook`` and
        ``skipkeys``.

    :param obj: a Python object
    :param fp: an open JSON file
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param check_circular: check for circular references
    :param commas: separate items by commas when indented
    :param end: the string to append at the end
    :param ensure_ascii: escape non-ASCII characters
    :param hook: the :ref:`hook <encoding_hook>` used for transforming data
    :param indent: the number of spaces or string to indent with
    :param indent_leaves: indent leaf objects and arrays
    :param max_indent_level: the level up to which to indent
    :param quoted_keys: quote keys which are :ref:`identifiers <identifiers>`
    :param separators: the item and key separator
    :param skipkeys: skip non-string keys
    :param sort_keys: sort the keys of objects
    :param trailing_comma: add a trailing comma when indented
    :param types: a dictionary of :ref:`additional types <protocol_types>`
    :raises RecursionError: if the object is too deeply nested
    :raises TypeError: for unserializable values
    :raises TruncatedSyntaxError: when failing to write to the file
    :raises ValueError: for invalid values

    Example:
        Writing to standard output:

        >>> import jsonyx as json
        >>> json.dump('"foo\bar')
        "\"foo\bar"
        >>> json.dump("\\")
        "\\"
        >>> json.dump("\u20AC")
        "€"

        Writing to an open file:

        >>> import jsonyx as json
        >>> from io import StringIO
        >>> io = StringIO()
        >>> json.dump(["streaming API"], io)
        >>> io.getvalue()
        '["streaming API"]\n'

    .. note:: The item separator is automatically stripped when indented.
    .. warning:: Avoid specifying ABCs for ``types``, that is very slow.

    """
    Encoder(
        allow=allow,
        check_circular=check_circular,
        commas=commas,
        end=end,
        ensure_ascii=ensure_ascii,
        hook=hook,
        indent=indent,
        indent_leaves=indent_leaves,
        max_indent_level=max_indent_level,
        quoted_keys=quoted_keys,
        separators=separators,
        skipkeys=skipkeys,
        sort_keys=sort_keys,
        trailing_comma=trailing_comma,
        types=types,
    ).dump(obj, fp)


def dumps(
    obj: object,
    *,
    allow: Container[str] = NOTHING,
    check_circular: bool = True,
    commas: bool = True,
    end: str = "\n",
    ensure_ascii: bool = False,
    hook: _Hook | None = None,
    indent: int | str | None = None,
    indent_leaves: bool = True,
    max_indent_level: int | None = None,
    quoted_keys: bool = True,
    separators: tuple[str, str] = (", ", ": "),
    skipkeys: bool = False,
    sort_keys: bool = False,
    trailing_comma: bool = False,
    types: dict[str, type | tuple[type, ...]] | None = None,
) -> str:
    r"""Serialize a Python object to a JSON string.

    .. versionchanged:: 2.0

        - Added ``commas``, ``indent_leaves``, ``max_indent_level``,
          ``quoted_keys`` and ``types``.
        - Made :class:`tuple` serializable by default instead of
          :class:`enum.Enum` and :class:`decimal.Decimal`.
        - Replaced ``item_separator`` and ``key_separator`` with
          ``separators``.

    .. versionchanged:: 2.1 Added ``check_circular``, ``hook`` and
        ``skipkeys``.

    :param obj: a Python object
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param check_circular: check for circular references
    :param commas: separate items by commas when indented
    :param end: the string to append at the end
    :param ensure_ascii: escape non-ASCII characters
    :param hook: the :ref:`hook <encoding_hook>` used for transforming data
    :param indent: the number of spaces or string to indent with
    :param indent_leaves: indent leaf objects and arrays
    :param max_indent_level: the level up to which to indent
    :param quoted_keys: quote keys which are :ref:`identifiers <identifiers>`
    :param separators: the item and key separator
    :param skipkeys: skip non-string keys
    :param sort_keys: sort the keys of objects
    :param trailing_comma: add a trailing comma when indented
    :param types: a dictionary of :ref:`additional types <protocol_types>`
    :raises RecursionError: if the object is too deeply nested
    :raises TypeError: for unserializable values
    :raises ValueError: for invalid values
    :return: a JSON string

    Example:
        >>> import jsonyx as json
        >>> json.dumps(["foo", {"bar": ("baz", None, 1.0, 2)}])
        '["foo", {"bar": ["baz", null, 1.0, 2]}]\n'

    .. note:: The item separator is automatically stripped when indented.
    .. warning:: Avoid specifying ABCs for ``types``, that is very slow.

    """
    return Encoder(
        allow=allow,
        check_circular=check_circular,
        commas=commas,
        end=end,
        ensure_ascii=ensure_ascii,
        hook=hook,
        indent=indent,
        indent_leaves=indent_leaves,
        max_indent_level=max_indent_level,
        quoted_keys=quoted_keys,
        separators=separators,
        skipkeys=skipkeys,
        sort_keys=sort_keys,
        trailing_comma=trailing_comma,
        types=types,
    ).dumps(obj)


def apply_patch(
    obj: Any,
    patch: _Operation | list[_Operation],
    *,
    allow: Container[str] = NOTHING,
    use_decimal: bool = False,
) -> Any:
    """Apply a JSON patch to a Python object.

    .. versionadded:: 2.0

    :param obj: a Python object
    :param patch: a :doc:`JSON patch </json-patch-spec>`
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param use_decimal: use :class:`decimal.Decimal` instead of :class:`float`
    :raises AssertionError: if an assertion fails
    :raises IndexError: if an index is out of range
    :raises JSONSyntaxError: if a query is invalid
    :raises KeyError: if a key is not found
    :raises TypeError: if a value has the wrong type
    :raises ValueError: if a value is invalid
    :return: the patched Python object

    Example:
        >>> import jsonyx as json
        >>> json.apply_patch([1, 2, 3], {"op": "del", "path": "$[1]"})
        [1, 3]

    .. tip:: Using a :ref:`filter` instead of an index is more robust.

    """
    return Manipulator(allow=allow, use_decimal=use_decimal).apply_patch(
        obj, patch,
    )


def paste_values(
    current_nodes: _Node | list[_Node],
    values: Any | list[Any],
    operation: _Operation,
    *,
    allow: Container[str] = NOTHING,
    use_decimal: bool = False,
) -> None:
    """Paste value to a node or values to a list of nodes.

    .. versionadded:: 2.0

    :param current_nodes: a node or a list of nodes
    :param values: a value or a list of values
    :param operation: a JSON :ref:`copy` or :ref:`move` operation
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param use_decimal: use :class:`decimal.Decimal` instead of :class:`float`
    :raises IndexError: if an index is out of range
    :raises JSONSyntaxError: if a query is invalid
    :raises KeyError: if a key is not found
    :raises TypeError: if a value has the wrong type
    :raises ValueError: if a value is invalid

    Example:
        >>> import jsonyx as json
        >>> obj = [1, 2, 3]
        >>> root = [obj]
        >>> node = root, 0  # pointer to obj
        >>> json.paste_values(node, 4, {"mode": "append"})
        >>> root[0]
        [1, 2, 3, 4]

    """
    Manipulator(allow=allow, use_decimal=use_decimal).paste_values(
        current_nodes, values, operation,
    )


def select_nodes(
    nodes: _Node | list[_Node],
    query: str,
    *,
    allow: Container[str] = NOTHING,
    allow_slice: bool = False,
    relative: bool = False,
    use_decimal: bool = False,
) -> list[_Node]:
    """Select nodes from a node or a list of nodes.

    .. versionadded:: 2.0

    :param nodes: a node or a list of nodes
    :param query: an :ref:`JSON select query <query>`
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param allow_slice: allow slice
    :param relative: query must start with ``"@"`` instead of ``"$"``
    :param use_decimal: use :class:`decimal.Decimal` instead of :class:`float`
    :raises IndexError: if an index is out of range
    :raises JSONSyntaxError: if the select query is invalid
    :raises KeyError: if a key is not found
    :raises TypeError: if a value has the wrong type
    :return: the selected list of nodes

    Example:
        >>> import jsonyx as json
        >>> obj = [1, 2, 3, 4, 5, 6]
        >>> root = [obj]
        >>> node = root, 0  # pointer to obj
        >>> for target, key in json.select_nodes(node, "$[@ > 3]"):
        ...     target[key] = None
        ...
        >>> root[0]
        [1, 2, 3, None, None, None]

    """
    return Manipulator(allow=allow, use_decimal=use_decimal).select_nodes(
        nodes, query, allow_slice=allow_slice, relative=relative,
    )


def apply_filter(
    nodes: _Node | list[_Node],
    query: str,
    *,
    allow: Container[str] = NOTHING,
    use_decimal: bool = False,
) -> list[_Node]:
    """Apply a JSON filter query to a node or a list of nodes.

    .. versionadded:: 2.0

    :param nodes: a node or a list of nodes
    :param query: a :ref:`JSON filter query <filter>`
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param use_decimal: use :class:`decimal.Decimal` instead of :class:`float`
    :raises IndexError: if an index is out of range
    :raises JSONSyntaxError: if the filter query is invalid
    :raises KeyError: if a key is not found
    :raises TypeError: if a value has the wrong type
    :return: the filtered list of nodes

    Example:
        >>> import jsonyx as json
        >>> obj = None
        >>> root = [obj]
        >>> node = root, 0  # pointer to obj
        >>> assert json.apply_filter(node, "@ == null")

    """
    return Manipulator(allow=allow, use_decimal=use_decimal).apply_filter(
        nodes, query,
    )


def load_query_value(
    s: str,
    *,
    allow: Container[str] = NOTHING,
    use_decimal: bool = False,
) -> Any:
    """Deserialize a JSON query value to a Python object.

    .. versionadded:: 2.0

    :param s: a :ref:`JSON query value <query_value>`
    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param use_decimal: use :class:`decimal.Decimal` instead of :class:`float`
    :raises JSONSyntaxError: if the query value is invalid
    :return: a Python object

    Example:
        >>> import jsonyx as json
        >>> json.load_query_value("'~'foo'")
        "'foo"

    """
    return Manipulator(allow=allow, use_decimal=use_decimal).load_query_value(
        s,
    )



"""JSON select_nodes tests."""
from __future__ import annotations

__all__: list[str] = []

from typing import TYPE_CHECKING, Any

import pytest

from jsonyx import JSONSyntaxError, select_nodes
from jsonyx.test import check_syntax_err

if TYPE_CHECKING:
    _Target = dict[Any, Any] | list[Any]
    _Node = tuple[_Target, Any]


# pylint: disable-next=R0903
class _Slicer:
    @staticmethod
    def __getitem__(item: slice) -> slice:
        return item


_slicer: _Slicer = _Slicer()


@pytest.mark.parametrize(("node", "keep"), [
    # List
    (([], _slicer[:]), True),
    (([], 0), False),
    (([0], 0), True),

    # Dict
    (({}, ""), False),
    (({"": 0}, ""), True),
])  # type: ignore
def test_optional_marker(node: _Node, keep: bool) -> None:
    """Test optional marker."""
    expected: list[_Node] = [node] if keep else []
    assert select_nodes(node, "$?", allow_slice=True) == expected


def test_optional_marker_not_allowed() -> None:
    """Test optional marker when not allowed."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        select_nodes([], "@?", relative=True)

    msg: str = "Optional markers are not allowed in relative query"
    check_syntax_err(exc_info, msg, 2, 3)


@pytest.mark.parametrize("key", [
    # First character
    "A", "_", "\u16ee", "\u1885", "\u2118",

    # Remaining characters
    "A0", "AA", "A_", "A\u0300", "A\u2118",
])
def test_property(key: str) -> None:
    """Test property."""
    assert select_nodes(([{}], 0), f"$.{key}") == [({}, key)]


@pytest.mark.parametrize("key", [
    # First character
    "\x00", "!", "$", "0", "\xb2", "\u0300", "\u037a", "\u0488",

    # Remaining characters
    "A\x00", "A!", "A$", "A\xb2", "A\u037a", "A\u0488",
])
def test_invalid_property(key: str) -> None:
    """Test invalid property."""
    with pytest.raises(JSONSyntaxError):
        select_nodes([], f"$.{key}")


@pytest.mark.parametrize("query", [
    # At the end
    "$.a",

    # In the middle
    "$.a?", "$.a.b", "$.a[:]", "$.a[0]", "$.a['']", "$.a[@]",
])
def test_list_property(query: str) -> None:
    """Test property on a list."""
    match: str = "List index must be int or slice, not"
    with pytest.raises(TypeError, match=match):
        select_nodes(([[]], 0), query, allow_slice=True)


@pytest.mark.parametrize("query", [
    # At the end
    "@.a",

    # In the middle
    "@.a.b", "@.a[0]", "@.a['']",
])
def test_relative_list_property(query: str) -> None:
    """Test relative property on a list."""
    with pytest.raises(TypeError, match="List index must be int, not"):
        select_nodes(([[]], 0), query, relative=True)


@pytest.mark.parametrize(("node", "keep"), [
    # List
    (([], 0), False),
    (([0], 0), True),

    # Dict
    (({}, ""), False),
    (({"": 0}, ""), True),
])  # type: ignore
def test_condition(node: _Node, keep: bool) -> None:
    """Test condition."""
    expected: list[_Node] = [node] if keep else []
    assert select_nodes(node, "${@}") == expected


@pytest.mark.parametrize("key", [
    # Ascending
    _slicer[:],

    # Descending
    _slicer[::-1],
])
def test_condition_slice(key: slice) -> None:
    """Test condition."""
    obj: list[Any] = [1, 2, 3]
    expected: list[_Node] = [(obj, 0), (obj, 1), (obj, 2)]
    assert select_nodes((obj, key), "${@}") == expected


def test_condition_not_allowed() -> None:
    """Test condition when not allowed."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        select_nodes([], "@{@}", relative=True)

    msg: str = "Conditions are not allowed in relative query"
    check_syntax_err(exc_info, msg, 3)


@pytest.mark.parametrize(("query", "expected"), [
    # Slice
    ("$[:]", _slicer[:]),
    ("$[-1:]", _slicer[-1:]),
    ("$[0:]", _slicer[0:]),
    ("$[1:]", _slicer[1:]),
    ("$[10:]", _slicer[10:]),
    ("$[11:]", _slicer[11:]),
    ("$[:-1]", _slicer[:-1]),
    ("$[:0]", _slicer[:0]),
    ("$[:1]", _slicer[:1]),
    ("$[:10]", _slicer[:10]),
    ("$[:11]", _slicer[:11]),

    # Extended slice
    ("$[::]", _slicer[::]),
    ("$[-1::]", _slicer[-1::]),
    ("$[0::]", _slicer[0::]),
    ("$[1::]", _slicer[1::]),
    ("$[10::]", _slicer[10::]),
    ("$[11::]", _slicer[11::]),
    ("$[:-1:]", _slicer[:-1:]),
    ("$[:0:]", _slicer[:0:]),
    ("$[:1:]", _slicer[:1:]),
    ("$[:10:]", _slicer[:10:]),
    ("$[:11:]", _slicer[:11:]),
    ("$[::-1]", _slicer[::-1]),
    ("$[::0]", _slicer[::0]),
    ("$[::1]", _slicer[::1]),
    ("$[::10]", _slicer[::10]),
    ("$[::11]", _slicer[::11]),
])
def test_slice(query: str, expected: slice) -> None:
    """Test slice."""
    node: _Node = [[]], 0
    assert select_nodes(node, query, allow_slice=True) == [([], expected)]


@pytest.mark.parametrize("query", [
    # Slice
    "$[1\uff10:]", "$[:1\uff10]",

    # Extended slice
    "$[1\uff10::]", "$[:1\uff10:]", "$[::1\uff10]",
])
def test_invalid_slice(query: str) -> None:
    """Test slice."""
    with pytest.raises(JSONSyntaxError):
        select_nodes([], query)


@pytest.mark.parametrize(("query", "msg", "colno"), [
    # Start
    ("$[{big_num}:]", "Start is too big", 3),
    ("$[{big_num}::]", "Start is too big", 3),

    # Stop
    ("$[:{big_num}]", "Stop is too big", 4),
    ("$[:{big_num}:]", "Stop is too big", 4),

    # Step
    ("$[::{big_num}]", "Step is too big", 5),
])
def test_too_big_slice_idx(
    big_num: str, query: str, msg: str, colno: int,
) -> None:
    """Test too big slice index."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        select_nodes([], query.format(big_num=big_num))

    check_syntax_err(exc_info, msg, colno, colno + len(big_num))


@pytest.mark.parametrize("query", [
    # At the end
    "@[:]",

    # In the middle
    "@[:].b", "@[:][:]", "@[:][0]", "@[:]['']",
])
def test_slice_not_allowed(query: str) -> None:
    """Test slice when not allowed."""
    with pytest.raises(TypeError, match="List index must be int, not"):
        select_nodes(([[]], 0), query, relative=True)


@pytest.mark.parametrize("query", [
    # At the end
    "$[:]",

    # In the middle
    "$[:]?", "$[:].b", "$[:][:]", "$[:][0]", "$[:]['']", "$[:][@]",
])
def test_dict_slice(query: str) -> None:
    """Test slice on a dict."""
    with pytest.raises(TypeError, match="Dict key must be str, not"):
        select_nodes(([{}], 0), query)


@pytest.mark.parametrize("num", [
    # Sign
    "-1",

    # Integer
    "0", "1", "10", "11",
])
def test_idx(num: str) -> None:
    """Test index."""
    assert select_nodes(([[]], 0), f"$[{num}]") == [([], int(num))]


def test_invalid_idx() -> None:
    """Test invalid idx."""
    with pytest.raises(JSONSyntaxError):
        select_nodes(([[]], 0), "$[1\uff10]")


def test_too_big_idx(big_num: str) -> None:
    """Test too big index."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        select_nodes([], f"$[{big_num}]")

    check_syntax_err(exc_info, "Index is too big", 3, 3 + len(big_num))


@pytest.mark.parametrize("query", [
    # At the end
    "$[0]",

    # In the middle
    "$[0]?", "$[0].b", "$[0][:]", "$[0][0]", "$[0]['']", "$[0][@]",
])
def test_dict_idx(query: str) -> None:
    """Test index on a dict."""
    with pytest.raises(TypeError, match="Dict key must be str, not"):
        select_nodes(([{}], 0), query)


@pytest.mark.parametrize(("obj", "query", "keys"), [
    ([1, 2, 3], "$[@]", [0, 1, 2]),
    ({"a": 1, "b": 2, "c": 3}, "$[@]", ["a", "b", "c"]),
])
def test_filter(obj: _Target, query: str, keys: list[Any]) -> None:
    """Test filter."""
    assert select_nodes(([obj], 0), query) == [(obj, key) for key in keys]


def test_filter_not_allowed() -> None:
    """Test filter when not allowed."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        select_nodes([], "@[@]", relative=True)

    check_syntax_err(exc_info, "Filters are not allowed in relative query", 3)


@pytest.mark.parametrize(("obj", "query", "expected"), [
    # Root
    (0, "$", ([0], 0)),

    # Key
    ({}, "$['']", ({}, "")),

    # Multiple levels
    ([[[0]]], "$[0][0][0]", ([0], 0)),
])  # type: ignore
def test_query(obj: _Target, query: str, expected: _Node) -> None:
    """Test root."""
    assert select_nodes(([obj], 0), query) == [expected]


@pytest.mark.parametrize(("query", "msg", "colno"), [
    ("", "Expecting an absolute query", 1),
    ("@", "Expecting an absolute query", 1),
    ("$[0", "Expecting a closing bracket", 4),
    ("$ $ $", "Expecting end of file", 2),
])
def test_invalid_query(query: str, msg: str, colno: int) -> None:
    """Test invalid query."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        select_nodes([], query)

    check_syntax_err(exc_info, msg, colno)


@pytest.mark.parametrize("query", ["", "$"])
def test_invalid_relative_query(query: str) -> None:
    """Test invalid relative query."""
    with pytest.raises(JSONSyntaxError) as exc_info:
        select_nodes([], query, relative=True)

    check_syntax_err(exc_info, "Expecting a relative query")


@pytest.mark.parametrize("query", [
    # At the end
    "$[0]",

    # In the middle
    "$[0]?", "$[0].b", "$[0][:]", "$[0][0]", "$[0]['']", "$[0][@]",
])
def test_invalid_target(query: str) -> None:
    """Test invalid target."""
    with pytest.raises(TypeError, match="Target must be dict or list, not"):
        select_nodes(([0], 0), query)



"""pytest fixtures."""
from __future__ import annotations

__all__: list[str] = ["get_big_num", "get_json"]

import sys
from typing import TYPE_CHECKING

import pytest

from jsonyx import JSONSyntaxError

if TYPE_CHECKING:
    from types import ModuleType


if sys.version_info >= (3, 10):
    from test.support.import_helper import import_fresh_module  # type: ignore
else:
    from test.support import import_fresh_module

pyjson: ModuleType | None = import_fresh_module("jsonyx", blocked=["_jsonyx"])
if cjson := import_fresh_module("jsonyx", fresh=["_jsonyx"]):
    # JSONSyntaxError is not cached inside the _jsonyx module
    cjson.JSONSyntaxError = JSONSyntaxError  # type: ignore


@pytest.fixture(name="big_num")
def get_big_num() -> str:
    """Get big number."""
    if not hasattr(sys, "get_int_max_str_digits"):
        pytest.skip("requires integer string conversion length limit")

    return "1" + "0" * sys.get_int_max_str_digits()


@pytest.fixture(params=[cjson, pyjson], ids=["cjson", "pyjson"], name="json")
def get_json(request: pytest.FixtureRequest) -> ModuleType:
    """Get JSON module."""
    json: ModuleType | None
    if (json := request.param) is None:
        pytest.xfail("module unavailable")
        # pylint: disable-next=W0101
        pytest.fail("module unavailable")

    return json  # type: ignore[no-any-return]



"""JSON test utils."""
from __future__ import annotations

__all__: list[str] = ["check_syntax_err"]

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import pytest


def check_syntax_err(
    exc_info: pytest.ExceptionInfo[Any],
    msg: str,
    colno: int = 1,
    end_colno: int = -1,
) -> None:
    """Check truncated syntax error."""
    exc: Any = exc_info.value
    if end_colno < 0:
        end_colno = colno

    assert exc.msg == msg
    assert exc.lineno == 1
    assert exc.end_lineno == 1
    assert exc.colno == colno
    assert exc.end_colno == end_colno



"""JSON loads tests."""
from __future__ import annotations

__all__: list[str] = []

from collections import UserString
from dataclasses import dataclass
from decimal import MAX_EMAX, Decimal
from math import isnan
from typing import TYPE_CHECKING, Any

import pytest

from jsonyx.allow import (
    COMMENTS, MISSING_COMMAS, NAN_AND_INFINITY, SURROGATES, TRAILING_COMMA,
    UNQUOTED_KEYS,
)
from jsonyx.test import check_syntax_err

if TYPE_CHECKING:
    from types import ModuleType


# pylint: disable-next=R0903
@dataclass
class _MyBool:
    value: Any

    def __bool__(self) -> bool:
        return bool(self.value)


@pytest.mark.parametrize(
    "string", ["\ud800", "\ud800$", "\udf48"],  # noqa: PT014
)
def test_surrogate(json: ModuleType, string: str) -> None:
    """Test surrogate."""
    b: bytes = f'"{string}"'.encode(errors="surrogatepass")
    assert json.loads(b, allow=SURROGATES) == string


@pytest.mark.parametrize(
    "string", ["\ud800", "\ud800$", "\udf48"],  # noqa: PT014
)
def test_surrogate_not_allowed(json: ModuleType, string: str) -> None:
    """Test surrogate when not allowed."""
    b: bytes = f'"{string}"'.encode(errors="surrogatepass")
    with pytest.raises(json.TruncatedSyntaxError):
        json.loads(b)


def test_utf8_bom(json: ModuleType) -> None:
    """Test UTF-8 BOM."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads("\ufeff0")

    check_syntax_err(exc_info, "Unexpected UTF-8 BOM", 1, 2)


def test_null(json: ModuleType) -> None:
    """Test null."""
    assert json.loads("null") is None


@pytest.mark.parametrize(("s", "expected"), [
    ("true", True),
    ("false", False),
])
@pytest.mark.parametrize("bool_type", [_MyBool, bool])
def test_bool(
    json: ModuleType, s: str, expected: bool, bool_type: type,
) -> None:
    """Test bool."""
    obj: object = json.loads(s, hooks={"bool": bool_type})
    assert isinstance(obj, bool_type)
    assert obj == bool_type(expected)


@pytest.mark.parametrize("s", ["NaN", "Infinity", "-Infinity"])
@pytest.mark.parametrize("float_type", [Decimal, float])
def test_nan_and_infinity(json: ModuleType, s: str, float_type: type) -> None:
    """Test NaN and infinity."""
    obj: object = json.loads(
        s, allow=NAN_AND_INFINITY, hooks={"float": float_type},
    )
    expected: Any = float_type(s)
    expected = float_type(expected)
    assert isinstance(obj, float_type)
    if isnan(expected):
        assert isnan(obj)  # type: ignore[arg-type]
    else:
        assert obj == expected


@pytest.mark.parametrize("s", ["NaN", "Infinity", "-Infinity"])
def test_nan_and_infinity_not_allowed(json: ModuleType, s: str) -> None:
    """Test NaN and infinity when not allowed."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(s)

    check_syntax_err(exc_info, f"{s} is not allowed", 1, len(s) + 1)


@pytest.mark.parametrize("s", [
    # Sign
    "-1",

    # Integer
    "0", "1", "10", "11",
])
@pytest.mark.parametrize("int_type", [Decimal, int])
def test_int(json: ModuleType, s: str, int_type: type) -> None:
    """Test integer."""
    obj: object = json.loads(s, hooks={"int": int_type})
    assert isinstance(obj, int_type)
    assert obj == int_type(s)


def test_too_big_int(json: ModuleType, big_num: str) -> None:
    """Test too big integer."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(big_num)

    check_syntax_err(exc_info, "Invalid number", 1, len(big_num) + 1)


def test_int_type(json: ModuleType) -> None:
    """Test int_type."""
    obj: object = json.loads("0", hooks={"int": Decimal})
    assert isinstance(obj, Decimal)
    assert obj == Decimal(0)


@pytest.mark.parametrize("s", [
    # Sign
    "-1.0",

    # Fraction
    "1.0", "1.01", "1.1", "1.11",

    # Fraction with trailing zeros
    "1.00", "1.10",

    # Exponent e
    "1E1",

    # Exponent sign
    "1e-1", "1e+1",

    # Exponent power
    "1e0", "1e1", "1e10", "1e11",

    # Exponent power with leading zeros
    "1e00", "1e01",

    # Parts
    "1.1e1", "-1e1", "-1.1", "-1.1e1",

    # Big exponent
    "1e400", "-1e400",
])
@pytest.mark.parametrize("float_type", [Decimal, float])
def test_rational_number(json: ModuleType, s: str, float_type: type) -> None:
    """Test rational number."""
    obj: object = json.loads(s, hooks={"float": float_type})
    assert isinstance(obj, float_type)
    assert obj == float_type(s)


@pytest.mark.parametrize("s", [f"1e{MAX_EMAX + 1}", f"-1e{MAX_EMAX + 1}"])
def test_too_big_number(json: ModuleType, s: str) -> None:
    """Test too big JSON number."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(s, hooks={"float": Decimal})

    check_syntax_err(exc_info, "Invalid number", 1, len(s) + 1)


@pytest.mark.parametrize("s", ["1\uff10", "0.\uff10", "0e\uff10"])
def test_invalid_number(json: ModuleType, s: str) -> None:
    """Test invalid number."""
    with pytest.raises(json.JSONSyntaxError):
        json.loads(s)


@pytest.mark.parametrize("s", [
    # Empty string
    "",

    # UTF-8
    "$", "\xa3", "\u0418", "\u0939", "\u20ac", "\ud55c", "\U00010348",
    "\U001096b3",

    # Surrogates
    "\ud800", "\ud800$", "\udf48",  # noqa: PT014

    # Multiple characters
    "foo",
])
def test_string(json: ModuleType, s: str) -> None:
    """Test JSON string."""
    assert json.loads(f'"{s}"') == s


@pytest.mark.parametrize(("s", "expected"), [
    # Backslash escapes
    (r"\"", '"'),
    (r"\\", "\\"),
    (r"\/", "/"),
    (r"\b", "\b"),
    (r"\f", "\f"),
    (r"\n", "\n"),
    (r"\r", "\r"),
    (r"\t", "\t"),

    # Unicode escape sequences
    (r"\u0024", "$"),
    (r"\u00a3", "\xa3"),
    (r"\u0418", "\u0418"),
    (r"\u0939", "\u0939"),
    (r"\u20ac", "\u20ac"),
    (r"\ud55c", "\ud55c"),
    (r"\ud800\udf48", "\U00010348"),
    (r"\udbe5\udeb3", "\U001096b3"),

    # Multiple characters
    (r"foo\\bar", r"foo\bar"),
])
def test_string_escapes(json: ModuleType, s: str, expected: str) -> None:
    """Test string escapes."""
    assert json.loads(f'"{s}"') == expected


@pytest.mark.parametrize(("s", "expected"), [
    (r"\ud800", "\ud800"),
    (r"\ud800\u0024", "\ud800$"),
    (r"\udf48", "\udf48"),
])
def test_surrogate_escape(json: ModuleType, s: str, expected: str) -> None:
    """Test surrogate escape."""
    assert json.loads(f'"{s}"', allow=SURROGATES) == expected


@pytest.mark.parametrize(("s", "msg", "colno", "end_colno"), [
    ('"foo', "Unterminated string", 1, 5),
    ('"foo\n', "Unterminated string", 1, 5),
    ('"foo\r', "Unterminated string", 1, 5),
    ('"foo\r\n', "Unterminated string", 1, 5),
    ('"\b"', "Unescaped control character", 2, 3),
    ('"\\', "Expecting escaped character", 3, -1),
    ('"\\\n', "Expecting escaped character", 3, -1),
    ('"\\\r', "Expecting escaped character", 3, -1),
    ('"\\\r\n', "Expecting escaped character", 3, -1),
    (r'"\a"', "Invalid backslash escape", 2, 4),
    (r'"\ud800"', "Surrogates are not allowed", 2, 8),
    (r'"\ud800\u0024"', "Surrogates are not allowed", 2, 8),
    (r'"\udf48"', "Surrogates are not allowed", 2, 8),
])
def test_invalid_string(
    json: ModuleType, s: str, msg: str, colno: int, end_colno: int,
) -> None:
    """Test invalid JSON string."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(s)

    check_syntax_err(exc_info, msg, colno, end_colno)


@pytest.mark.parametrize(("s", "colno", "end_colno"), [
    # Single
    (r'"\u"', 4, 5),
    (r'"\u0xff"', 4, 8),  # Hex prefix
    # pylint: disable-next=C0209
    (r'"\u{}"'.format("\uff10" * 4), 4, 8),  # Unicode digits
    (r'"\u 000"', 4, 8),  # Surrounded by whitespace
    (r'"\u-000"', 4, 8),  # Negative number
    (r'"\u+000"', 4, 8),  # Positive number
    (r'"\u0_00"', 4, 8),  # Underscore between digits
    (r'"\u????"', 4, 8),

    # After high surrogate
    (r'"\ud800\u"', 10, 11),
    (r'"\ud800\u0xff"', 10, 14),  # Hex prefix
    # pylint: disable-next=C0209
    (r'"\ud800\u{}"'.format("\uff10" * 4), 10, 14),  # Unicode digits
    (r'"\ud800\u 000"', 10, 14),  # Surrounded by whitespace
    (r'"\ud800\u-000"', 10, 14),  # Negative number
    (r'"\ud800\u+000"', 10, 14),  # Positive number
    (r'"\ud800\u0_00"', 10, 14),  # Underscore between digits
    (r'"\ud800\u????"', 10, 14),
])
def test_invalid_unicode_escape(
    json: ModuleType, s: str, colno: int, end_colno: int,
) -> None:
    """Test invalid unicode escape."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(s)

    check_syntax_err(exc_info, "Expecting 4 hex digits", colno, end_colno)


def test_str_type(json: ModuleType) -> None:
    """Test str_type."""
    obj: object = json.loads('""', hooks={"str": UserString})
    assert isinstance(obj, UserString)
    assert not obj


@pytest.mark.parametrize(("s", "expected"), [
    # Empty array
    ("[]", []),

    # One value
    ("[0]", [0]),

    # Multiple values
    ("[1, 2, 3]", [1, 2, 3]),
])
@pytest.mark.parametrize("array_type", [list, tuple])
def test_array(
    json: ModuleType, s: str, expected: list[object], array_type: type,
) -> None:
    """Test JSON array."""
    types: dict[str, type] = {"array": array_type}
    assert json.loads(s, hooks=types) == array_type(expected)


@pytest.mark.parametrize(("s", "expected"), [
    # No whitespace
    ("[]", []),
    ("[1,2,3]", [1, 2, 3]),

    # In empty array
    ("[ ]", []),

    # Before first element
    ("[ 1,2,3]", [1, 2, 3]),

    # Before commas
    ("[1 ,2 ,3]", [1, 2, 3]),

    # After commas
    ("[1, 2, 3]", [1, 2, 3]),

    # After last element
    ("[1,2,3 ]", [1, 2, 3]),
])
def test_array_whitespace(
    json: ModuleType, s: str, expected: list[object],
) -> None:
    """Test whitespace in JSON array."""
    assert json.loads(s) == expected


@pytest.mark.parametrize(("s", "expected"), [
    # In empty array
    ("[/**/]", []),

    # Before first element
    ("[/**/1,2,3]", [1, 2, 3]),

    # Before commas
    ("[1/**/,2/**/,3]", [1, 2, 3]),

    # After commas
    ("[1,/**/2,/**/3]", [1, 2, 3]),

    # After last element
    ("[1,2,3/**/]", [1, 2, 3]),
])
def test_array_comments(
    json: ModuleType, s: str, expected: list[object],
) -> None:
    """Test comments in JSON array."""
    assert json.loads(s, allow=COMMENTS) == expected


@pytest.mark.parametrize(("s", "msg", "colno", "end_colno"), [
    ("[", "Unterminated array", 1, 2),
    ("[0", "Unterminated array", 1, 3),
    ("[1-2-3]", "Expecting comma", 3, -1),
    ("[1 2 3]", "Missing commas are not allowed", 3, -1),
    ("[0,", "Unterminated array", 1, 4),
    ("[0,]", "Trailing comma is not allowed", 3, 4),
])
def test_invalid_array(
    json: ModuleType, s: str, msg: str, colno: int, end_colno: int,
) -> None:
    """Test invalid JSON array."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(s)

    check_syntax_err(exc_info, msg, colno, end_colno)


@pytest.mark.parametrize(("s", "expected"), [
    # Empty object
    ("{}", []),

    # One value
    ('{"": 0}', [("", 0)]),

    # Multiple values
    ('{"a": 1, "b": 2, "c": 3}', [("a", 1), ("b", 2), ("c", 3)]),

    # Duplicate keys
    ('{"a": 1, "a": 2, "a": 3}', [("a", 1), ("a", 2), ("a", 3)]),
])
@pytest.mark.parametrize("object_type", [dict, list])
def test_object(
    json: ModuleType, s: str, expected: list[tuple[str, object]],
    object_type: type,
) -> None:
    """Test JSON object."""
    types: dict[str, type] = {"object": object_type}
    assert json.loads(s, hooks=types) == object_type(expected)


@pytest.mark.parametrize(("s", "expected"), [
    # No whitespace
    ("{}", {}),
    ('{"a":1,"b":2,"c":3}', {"a": 1, "b": 2, "c": 3}),

    # In empty object
    ("{ }", {}),

    # Before first element
    ('{ "a":1,"b":2,"c":3}', {"a": 1, "b": 2, "c": 3}),

    # Before colon
    ('{"a" :1,"b" :2,"c" :3}', {"a": 1, "b": 2, "c": 3}),

    # After colon
    ('{"a": 1,"b": 2,"c": 3}', {"a": 1, "b": 2, "c": 3}),

    # Before comma
    ('{"a":1 ,"b":2 ,"c":3}', {"a": 1, "b": 2, "c": 3}),

    # After comma
    ('{"a":1, "b":2, "c":3}', {"a": 1, "b": 2, "c": 3}),

    # After last element
    ('{"a":1,"b":2,"c":3 }', {"a": 1, "b": 2, "c": 3}),
])
def test_object_whitespace(
    json: ModuleType, s: str, expected: dict[str, object],
) -> None:
    """Test whitespace in JSON object."""
    assert json.loads(s) == expected


@pytest.mark.parametrize(("s", "expected"), [
    # In empty object
    ("{/**/}", {}),

    # Before first element
    ('{/**/"a":1,"b":2,"c":3}', {"a": 1, "b": 2, "c": 3}),

    # Before colon
    ('{"a"/**/:1,"b"/**/:2,"c"/**/:3}', {"a": 1, "b": 2, "c": 3}),

    # After colon
    ('{"a":/**/1,"b":/**/2,"c":/**/3}', {"a": 1, "b": 2, "c": 3}),

    # Before comma
    ('{"a":1/**/,"b":2/**/,"c":3}', {"a": 1, "b": 2, "c": 3}),

    # After comma
    ('{"a":1,/**/"b":2,/**/"c":3}', {"a": 1, "b": 2, "c": 3}),

    # After last element
    ('{"a":1,"b":2,"c":3/**/}', {"a": 1, "b": 2, "c": 3}),
])
def test_object_comments(
    json: ModuleType, s: str, expected: dict[str, object],
) -> None:
    """Test comments in JSON object."""
    assert json.loads(s, allow=COMMENTS) == expected


@pytest.mark.parametrize("key", [
    # First character
    "A", "_", "\u16ee", "\u1885", "\u2118",

    # Remaining characters
    "A0", "AA", "A_", "A\u0300", "A\u2118",
])
def test_unquoted_key(json: ModuleType, key: str) -> None:
    """Test unquoted key."""
    assert json.loads(f"{{{key}: 0}}", allow=UNQUOTED_KEYS) == {key: 0}


@pytest.mark.parametrize("key", [
    # First character
    "A", "_", "\u16ee", "\u1885", "\u2118",

    # Remaining characters
    "A0", "AA", "A_", "A\u0300", "A\u2118",
])
def test_unquoted_key_not_allowed(json: ModuleType, key: str) -> None:
    """Test unquoted key when not allowed."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(f"{{{key}: 0}}")

    msg: str = "Unquoted keys are not allowed"
    check_syntax_err(exc_info, msg, 2, 2 + len(key))


@pytest.mark.parametrize("key", [
    # First character
    "\x00", "!", "$", "0", "\xb2", "\u0300", "\u037a", "\u0488",

    # Remaining characters
    "A\x00", "A!", "A$", "A\xb2", "A\u037a", "A\u0488",
])
def test_invalid_unquoted_key(json: ModuleType, key: str) -> None:
    """Test invalid unquoted key."""
    with pytest.raises(json.JSONSyntaxError):
        json.loads(f"{{{key}: 0}}", allow=UNQUOTED_KEYS)


@pytest.mark.parametrize(("s", "msg", "colno", "end_colno"), [
    ("{", "Unterminated object", 1, 2),
    ('{""}', "Expecting colon", 4, -1),
    ('{"": 0', "Unterminated object", 1, 7),
    ('{"a": 1"b": 2"c": 3}', "Expecting comma", 8, -1),
    ('{"a": 1 "b": 2 "c": 3}', "Missing commas are not allowed", 8, -1),
    ('{"": 0,', "Unterminated object", 1, 8),
    ('{"": 0,}', "Trailing comma is not allowed", 7, 8),
])
def test_invalid_object(
    json: ModuleType, s: str, msg: str, colno: int, end_colno: int,
) -> None:
    """Test invalid JSON object."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(s)

    check_syntax_err(exc_info, msg, colno, end_colno)


def test_repeated_keys(json: ModuleType) -> None:
    """Test repeated keys."""
    assert id(b"abc".decode()) != id("abc")
    s: str = '[{"abc": 1}, {"abc": 2}, {"abc": 3}]'
    dcts: list[dict[str, int]] = json.loads(s)
    assert len({id(key) for dct in dcts for key in dct}) == 3


def test_repeated_keys_cache_keys(json: ModuleType) -> None:
    """Test repeated keys with cache_keys."""
    assert id(b"abc".decode()) != id("abc")
    s: str = '[{"abc": 1}, {"abc": 2}, {"abc": 3}]'
    dcts: list[dict[str, int]] = json.loads(s, cache_keys=True)
    assert len({id(key) for dct in dcts for key in dct}) == 1


@pytest.mark.parametrize("s", ["", "-", "foo"])
def test_expecting_value(json: ModuleType, s: str) -> None:
    """Test expecting JSON value."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(s)

    check_syntax_err(exc_info, "Expecting value")


@pytest.mark.parametrize(("s", "expected"), [
    ("[1 2 3]", [1, 2, 3]),
    ('{"a": 1 "b": 2 "c": 3}', {"a": 1, "b": 2, "c": 3}),
])
def test_missing_comma(
    json: ModuleType, s: str, expected: dict[str, object] | list[object],
) -> None:
    """Test missing comma."""
    assert json.loads(s, allow=MISSING_COMMAS) == expected


@pytest.mark.parametrize(("s", "expected"), [
    ("[0,]", [0]),
    ('{"": 0,}', {"": 0}),
])
def test_trailing_comma(
    json: ModuleType, s: str, expected: dict[str, object] | list[object],
) -> None:
    """Test trailing comma."""
    assert json.loads(s, allow=TRAILING_COMMA) == expected


@pytest.mark.parametrize(("start", "msg"), [
    ("[", "Array is too deeply nested"),
    ('{"":', "Object is too deeply nested"),
])
def test_recursion(json: ModuleType, start: str, msg: str) -> None:
    """Test recursion."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(start * 100_000)

    assert exc_info.value.msg == msg


@pytest.mark.parametrize("s", [
    # No whitespace
    "0",

    # Before value
    " 0",

    # After value
    "0 ",
])
def test_value_whitespace(json: ModuleType, s: str) -> None:
    """Test whitespace around JSON value."""
    assert json.loads(s) == 0


@pytest.mark.parametrize("s", [
    # Before value
    "/**/0",

    # After value
    "0/**/",
])
def test_value_comments(json: ModuleType, s: str) -> None:
    """Test comments around JSON value."""
    assert json.loads(s, allow=COMMENTS) == 0


@pytest.mark.parametrize("s", [
    # One character
    "0 ", "0\t", "0\n", "0\r",

    # Multiple characters
    "0   ",
])
def test_whitespace(json: ModuleType, s: str) -> None:
    """Test whitespace."""
    assert json.loads(s) == 0


@pytest.mark.parametrize("s", [
    # One comment
    "//\n0", "//\r0", "//\r\n0",
    "//line comment\n0", "//line comment\r0", "//line comment\r\n0",
    "/*block comment*/0",

    # Multiple comments
    "//comment 1\n//comment 2\n//comment 3\n0",
    "//comment 1\r//comment 2\r//comment 3\n0",
    "//comment 1\r\n//comment 2\r\n//comment 3\r\n0",
    "/*comment 1*//*comment 2*//*comment 3*/0",

    # Whitespace
    " //comment 1\n //comment 2\n //comment 3\n 0",
    " //comment 1\r //comment 2\r //comment 3\r 0",
    " //comment 1\r\n //comment 2\r\n //comment 3\r\n 0",
    " /*comment 1*/ /*comment 2*/ /*comment 3*/ 0",
])
def test_comments(json: ModuleType, s: str) -> None:
    """Test comments."""
    assert json.loads(s, allow=COMMENTS) == 0


def test_invalid_comment(json: ModuleType) -> None:
    """Test invalid comment."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads("/*unterminated comment", allow=COMMENTS)

    check_syntax_err(exc_info, "Unterminated comment", 1, 23)


@pytest.mark.parametrize(("s", "end_colno"), [
    ("//line comment\n0", 15),
    ("//line comment\r0", 15),
    ("//line comment\r\n0", 15),
    ("/*block comment*/0", 18),
    ("/*unterminated comment", 23),
])
def test_comments_not_allowed(
    json: ModuleType, s: str, end_colno: int,
) -> None:
    """Test comments when not allowed."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads(s)

    check_syntax_err(exc_info, "Comments are not allowed", 1, end_colno)


def test_end_of_file(json: ModuleType) -> None:
    """Test end of file."""
    with pytest.raises(json.JSONSyntaxError) as exc_info:
        json.loads("1 2 3")

    check_syntax_err(exc_info, "Expecting end of file", 3)



"""JSON tests."""
from __future__ import annotations

__all__: list[str] = []

from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING

import pytest

from jsonyx import JSONSyntaxError, detect_encoding, format_syntax_error

if TYPE_CHECKING:
    from types import ModuleType


@pytest.mark.parametrize(("s", "encoding"), [
    # JSON must start with ASCII character (not NULL)
    # Strings can't contain control characters (including NULL)

    # Empty string
    ("", "utf-8"),

    # utf-8
    ("XX", "utf-8"),  # 1 ASCII character
    ("XX XX", "utf-8"),  # 2 ASCII characters
    ("XX XX XX", "utf-8"),  # 3 ASCII characters
    ("XX XX XX XX", "utf-8"),  # 4 ASCII characters

    # utf-8-sig
    ("ef bb bf", "utf-8-sig"),  # Empty JSON with BOM
    ("ef bb bf XX", "utf-8-sig"),  # BOM + 1 ASCII character

    # utf-16-be
    ("00 XX", "utf-16-be"),  # 1 ASCII character (BE)
    ("00 XX 00 XX", "utf-16-be"),  # 2 ASCII characters (BE)
    ("00 XX XX 00", "utf-16-be"),  # 1 ASCII + 1 Unicode character (BE)
    ("00 XX XX XX", "utf-16-be"),  # 1 ASCII + 1 Unicode character (BE)

    # utf-16-le
    ("XX 00", "utf-16-le"),  # 1 ASCII character (LE)
    ("XX 00 00 XX", "utf-16-le"),  # 1 ASCII + 1 Unicode character (LE)
    ("XX 00 XX 00", "utf-16-le"),  # 2 ASCII characters (LE)
    ("XX 00 XX XX", "utf-16-le"),  # 1 ASCII + 1 Unicode character (LE)

    # utf-16
    ("fe ff", "utf-16"),  # Empty JSON with BOM (BE)
    ("fe ff 00 XX", "utf-16"),  # BOM + 1 ASCII character (BE)
    ("ff fe", "utf-16"),  # Empty JSON with BOM (LE)
    ("ff fe XX 00", "utf-16"),  # BOM + 1 ASCII character (LE)

    # utf-32-be
    ("00 00 00 XX", "utf-32-be"),  # 1 ASCII character (BE)

    # utf-32-le
    ("XX 00 00 00", "utf-32-le"),  # 1 ASCII character (LE)

    # utf-32
    ("00 00 fe ff", "utf-32"),  # Empty JSON with BOM (BE)
    ("00 00 fe ff 00 00 00 XX", "utf-32"),  # BOM + 1 ASCII character (BE)
    ("ff fe 00 00", "utf-32"),  # Empty JSON with BOM (LE)
    ("ff fe 00 00 XX 00 00 00", "utf-32"),  # BOM + 1 ASCII character (LE)
])
def test_detect_encoding(s: str, encoding: str) -> None:
    """Test detect JSON encoding."""
    assert detect_encoding(bytes.fromhex(s.replace("XX", "01"))) == encoding


@pytest.mark.parametrize(("doc", "end", "line_range", "column_range"), [
    ("line 1", 5, "1", "6"),
    #      ^
    ("line 1", 6, "1", "6-7"),
    #      ^
    ("line 1\nline 2", 12, "1-2", "6"),
    #      ^^^^^^^^
    ("line 1\nline 2", 13, "1-2", "6-7"),
    #      ^^^^^^^^^
])
def test_format_syntax_error(
    doc: str, end: int, line_range: str, column_range: str,
) -> None:
    """Test format_syntax_error."""
    exc: JSONSyntaxError = JSONSyntaxError("", "<string>", doc, 5, end)
    assert format_syntax_error(exc)[0] == (
        f'  File "<string>", line {line_range}, column {column_range}\n'
    )


def test_load(json: ModuleType) -> None:
    """Test JSON load."""
    assert json.load(StringIO("0")) == 0


def test_read(json: ModuleType) -> None:
    """Test JSON read."""
    with TemporaryDirectory() as tmpdir:
        filename: Path = Path(tmpdir) / "file.json"
        filename.write_text("0", "utf-8")
        assert json.read(filename) == 0



"""JSON differ."""
from __future__ import annotations

__all__: list[str] = ["make_patch"]

import re
from decimal import Decimal
from math import isnan
from re import DOTALL, MULTILINE, VERBOSE, Match
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Callable, KeysView

    _Operation = dict[str, Any]
    _SubFunc = Callable[[str | Callable[[Match[str]], str], str], str]

_REPL: str = r"~\1"

_escape: _SubFunc = re.compile(r"(['~])", VERBOSE | MULTILINE | DOTALL).sub


def _encode_query_key(key: str) -> str:
    return f".{key}" if key.isidentifier() else f"['{_escape(_REPL, key)}']"


def _eq(a: Any, b: Any) -> bool:
    if type(a) is not type(b):
        result = False
    elif isinstance(a, dict):
        if a.keys() != b.keys():
            result = False
        else:
            result = all(_eq(a[key], b[key]) for key in a)  # type: ignore
    elif isinstance(a, list):
        if len(a) != len(b):  # type: ignore
            result = False
        else:
            result = all(map(_eq, a, b))  # type: ignore
    elif isinstance(a, (Decimal, float)) and isnan(a):
        result = isnan(b)
    else:
        result = a == b

    return result


def _get_lcs(old: list[Any], new: list[Any]) -> list[Any]:
    dp: list[list[int]] = [[0] * (len(new) + 1) for _ in range(len(old) + 1)]
    for i, old_value in enumerate(old):
        for j, new_value in enumerate(new):
            if _eq(old_value, new_value):
                dp[i + 1][j + 1] = dp[i][j] + 1
            else:
                dp[i + 1][j + 1] = max(dp[i + 1][j], dp[i][j + 1])

    lcs: list[Any] = []
    i, j = len(old), len(new)
    while i > 0 and j > 0:
        if _eq(old[i - 1], new[j - 1]):
            lcs.append(old[i - 1])
            i -= 1
            j -= 1
        elif dp[i - 1][j] >= dp[i][j - 1]:
            i -= 1
        else:
            j -= 1

    return lcs[::-1]


def _make_patch(
    old: Any, new: Any, patch: list[_Operation], path: str = "$",
) -> None:
    if _eq(old, new):
        return

    if isinstance(old, dict) and isinstance(new, dict):
        old_keys: KeysView[Any] = old.keys()  # type: ignore
        new_keys: KeysView[Any] = new.keys()  # type: ignore
        for key in old_keys:  # type: ignore
            new_path: str = f"{path}{_encode_query_key(key)}"  # type: ignore
            if key in new_keys:
                _make_patch(old[key], new[key], patch, new_path)
            else:
                patch.append({"op": "del", "path": new_path})

        for key in new_keys:  # type: ignore
            if key not in old_keys:  # preserve order
                new_path = f"{path}{_encode_query_key(key)}"  # type: ignore
                patch.append(
                    {"op": "set", "path": new_path, "value": new[key]},
                )
    elif isinstance(old, list) and isinstance(new, list):
        lcs: list[Any] = _get_lcs(old, new)  # type: ignore
        old_idx = new_idx = lcs_idx = 0
        while old_idx < len(old) or new_idx < len(new):  # type: ignore
            new_path = f"{path}[{new_idx}]"
            removed: bool = old_idx < len(old) and (  # type: ignore
                lcs_idx >= len(lcs) or not _eq(old[old_idx], lcs[lcs_idx])
            )
            inserted: bool = new_idx < len(new) and (  # type: ignore
                lcs_idx >= len(lcs) or not _eq(new[new_idx], lcs[lcs_idx])
            )
            if removed and inserted:
                _make_patch(old[old_idx], new[new_idx], patch, new_path)
                old_idx += 1
                new_idx += 1
            elif removed:
                patch.append({"op": "del", "path": new_path})
                old_idx += 1
            elif inserted:
                patch.append(
                    {"op": "insert", "path": new_path, "value": new[new_idx]},
                )
                new_idx += 1
            else:
                old_idx += 1
                new_idx += 1
                lcs_idx += 1
    else:
        patch.append({"op": "set", "path": path, "value": new})


def make_patch(old: Any, new: Any) -> list[_Operation]:
    """Make a JSON patch from two Python objects.

    .. versionadded:: 2.0

    :param old: the old Python object
    :param new: the new Python object
    :raises RecursionError: if the object is too deeply nested
    :return: the :doc:`JSON patch </json-patch-spec>`

    Example:
        >>> import jsonyx as json
        >>> json.make_patch([1, 2, 3], [1, 3])
        [{'op': 'del', 'path': '$[1]'}]

    """
    patch: list[_Operation] = []
    _make_patch(old, new, patch)
    return patch


make_patch.__module__ = "jsonyx"



"""JSON apply_patch tests."""
from __future__ import annotations

__all__: list[str] = []

from re import escape
from typing import TYPE_CHECKING, Any

import pytest

from jsonyx import apply_patch

if TYPE_CHECKING:
    _Operation = dict[str, Any]


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    ([1, 2, 3], {}, [1, 2, 3, 4]),
    ([[1, 2, 3]], {"path": "$[0]"}, [[1, 2, 3, 4]]),
])
def test_append(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test append."""
    assert apply_patch(obj, {"op": "append", "value": 4, **kwargs}) == expected


def test_append_copy() -> None:
    """Test if append makes a copy."""
    value: list[Any] = [0]
    result: list[Any] = apply_patch([], {"op": "append", "value": value})[0]
    assert result == value  # sanity check
    assert result is not value


@pytest.mark.parametrize(("obj", "kwargs"), [
    (0, {}),
    ([0], {"path": "$[0]"}),
])
def test_assert(obj: Any, kwargs: _Operation) -> None:
    """Test assert."""
    apply_patch(obj, {"op": "assert", "expr": "@ == 0", **kwargs})


@pytest.mark.parametrize(("obj", "kwargs", "match"), [
    (1, {}, "Path $: @ == 0"),
    ([1], {"path": "$[0]"}, "Path $[0]: @ == 0"),
    (1, {"msg": "msg"}, "msg"),
])
def test_failed_assert(obj: Any, kwargs: _Operation, match: str) -> None:
    """Test failed assert."""
    with pytest.raises(AssertionError, match=escape(match)):
        apply_patch(obj, {"op": "assert", "expr": "@ == 0", **kwargs})


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    # Without path
    ([1, 2, 3], {}, []),
    ({"a": 1, "b": 2, "c": 3}, {}, {}),

    # With path
    ([[1, 2, 3]], {"path": "$[0]"}, [[]]),
    ([{"a": 1, "b": 2, "c": 3}], {"path": "$[0]"}, [{}]),
])  # type: ignore
def test_clear(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test clear."""
    assert apply_patch(obj, {"op": "clear", **kwargs}) == expected


def test_invalid_clear() -> None:
    """Test invalid clear."""
    with pytest.raises(TypeError, match="Target must be dict or list, not"):
        apply_patch(0, {"op": "clear"})


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    # Without path
    ([0], {"mode": "append", "from": "@[0]"}, [0, 0]),
    ({"a": 0}, {"mode": "set", "from": "@.a", "to": "@.b"}, {"a": 0, "b": 0}),

    # With path
    ([[0]], {"mode": "append", "path": "$[0]", "from": "@[0]"}, [[0, 0]]),
    (
        [{"a": 0}],
        {"mode": "set", "path": "$[0]", "from": "@.a", "to": "@.b"},
        [{"a": 0, "b": 0}],
    ),

    # Allow slice
    ([1, 2, 3], {"mode": "extend", "from": "@[::-1]"}, [1, 2, 3, 3, 2, 1]),
])
def test_copy(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test copy."""
    assert apply_patch(obj, {"op": "copy", **kwargs}) == expected


def test_copy_copy() -> None:
    """Test if copy makes a copy."""
    value: list[Any] = [0]
    patch: _Operation = {"op": "copy", "mode": "set", "from": "@[0]"}
    result: list[Any] = apply_patch([value], patch)
    assert result == value  # sanity check
    assert result is not value


@pytest.mark.parametrize(("obj", "path", "expected"), [
    # List and dict
    ([1, 2, 3, 4], "$[3]", [1, 2, 3]),
    ({"a": 1, "b": 2, "c": 3, "d": 4}, "$.d", {"a": 1, "b": 2, "c": 3}),

    # Allow slice
    ([1, 2, 3], "$[:]", []),

    # Reverse indices for queries
    ([1, 0, 2, 0, 3], "$[@ == 0]", [1, 2, 3]),
])
def test_del(obj: Any, path: str, expected: Any) -> None:
    """Test delete."""
    assert apply_patch(obj, {"op": "del", "path": path}) == expected


def test_del_root() -> None:
    """Test delete root."""
    with pytest.raises(ValueError, match="Can not delete root"):
        apply_patch(0, {"op": "del", "path": "$"})


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    ([1, 2, 3], {}, [1, 2, 3, 4, 5, 6]),
    ([[1, 2, 3]], {"path": "$[0]"}, [[1, 2, 3, 4, 5, 6]]),
])
def test_extend(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test extend."""
    patch: _Operation = {"op": "extend", "values": [4, 5, 6], **kwargs}
    assert apply_patch(obj, patch) == expected


def test_extend_copy() -> None:
    """Test if extend makes a copy."""
    value: list[Any] = [0]
    result: list[Any] = apply_patch([], {"op": "extend", "values": [value]})[0]
    assert result == value  # sanity check
    assert result is not value


@pytest.mark.parametrize(("path", "expected"), [
    # Normal
    ("$[0]", [0, 1, 2, 3]),

    # Reverse indices for queries
    ("$[@]", [0, 1, 0, 2, 0, 3]),
])
def test_insert(path: str, expected: Any) -> None:
    """Test insert."""
    patch: _Operation = {"op": "insert", "path": path, "value": 0}
    assert apply_patch([1, 2, 3], patch) == expected


def test_insert_copy() -> None:
    """Test if insert makes a copy."""
    value: list[Any] = [0]
    patch: _Operation = {"op": "insert", "path": "$[0]", "value": value}
    result: list[Any] = apply_patch([], patch)[0]
    assert result == value  # sanity check
    assert result is not value


def test_insert_root() -> None:
    """Test insert at root."""
    with pytest.raises(ValueError, match="Can not insert at root"):
        apply_patch(0, {"op": "insert", "path": "$", "value": 0})


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    # Without path
    ([1, 2], {"mode": "append", "from": "@[0]"}, [2, 1]),
    ({"a": 0}, {"mode": "set", "from": "@.a", "to": "@.b"}, {"b": 0}),

    # With path
    ([[1, 2]], {"mode": "append", "path": "$[0]", "from": "@[0]"}, [[2, 1]]),
    (
        [{"a": 0}],
        {"mode": "set", "path": "$[0]", "from": "@.a", "to": "@.b"},
        [{"b": 0}],
    ),

    # Allow slice
    ([1, 2, 3], {"mode": "extend", "from": "@[::-1]"}, [3, 2, 1]),
])
def test_move(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test move."""
    assert apply_patch(obj, {"op": "move", **kwargs}) == expected


def test_move_current_object() -> None:
    """Test move current object."""
    with pytest.raises(ValueError, match="Can not move current object"):
        apply_patch(0, {"op": "move", "mode": "set", "from": "@"})


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    ([1, 2, 3], {}, [3, 2, 1]),
    ([[1, 2, 3]], {"path": "$[0]"}, [[3, 2, 1]]),
])
def test_reverse(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test reverse."""
    assert apply_patch(obj, {"op": "reverse", **kwargs}) == expected


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    # Normal
    (0, {"value": 1}, 1),
    ([0], {"path": "$[0]", "value": 1}, [1]),
    ({"a": 0}, {"path": "$.a", "value": 1}, {"a": 1}),

    # Allow slice
    ([1, 2, 3], {"path": "$[:]", "value": [4, 5, 6]}, [4, 5, 6]),
])
def test_set(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test set."""
    assert apply_patch(obj, {"op": "set", **kwargs}) == expected


def test_set_copy() -> None:
    """Test if set makes a copy."""
    value: list[Any] = [1]
    result: list[Any] = apply_patch([0], {"op": "set", "value": value})
    assert result == value  # sanity check
    assert result is not value


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    ([3, 1, 2], {}, [1, 2, 3]),
    ([[3, 1, 2]], {"path": "$[0]"}, [[1, 2, 3]]),
    ([3, 1, 2], {"reverse": True}, [3, 2, 1]),
])
def test_sort(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test sort."""
    assert apply_patch(obj, {"op": "sort", **kwargs}) == expected


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    ({"a": 1, "b": 2, "c": 3}, {}, {"a": 4, "b": 5, "c": 6}),
    ([{"a": 1, "b": 2, "c": 3}], {"path": "$[0]"}, [{"a": 4, "b": 5, "c": 6}]),
])
def test_update(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test update."""
    properties: dict[str, Any] = {"a": 4, "b": 5, "c": 6}
    patch: _Operation = {"op": "update", "properties": properties, **kwargs}
    assert apply_patch(obj, patch) == expected


def test_update_copy() -> None:
    """Test if update makes a copy."""
    value: list[Any] = [1]
    patch: _Operation = {"op": "update", "properties": {"a": [1]}}
    result: list[Any] = apply_patch({}, patch)["a"]
    assert result == value  # sanity check
    assert result is not value


@pytest.mark.parametrize(("paths", "expected"), [
    # Empty list
    ([], {"a": 1, "b": 2, "c": 3}),

    # Multiple operations
    (["$.a", "$.b", "$.c"], {}),
])
def test_operations(paths: list[str], expected: dict[str, Any]) -> None:
    """Test operations."""
    patch: list[_Operation] = [{"op": "del", "path": path} for path in paths]
    assert apply_patch({"a": 1, "b": 2, "c": 3}, patch) == expected


def test_unknown_operation() -> None:
    """Test unknown operation."""
    with pytest.raises(ValueError, match="Unknown operation"):
        apply_patch(0, {"op": "foo"})



"""JSON decoder."""
from __future__ import annotations

__all__: list[str] = [
    "Decoder", "JSONSyntaxError", "TruncatedSyntaxError", "detect_encoding",
]

import re
import sys
from codecs import (
    BOM_UTF8, BOM_UTF16_BE, BOM_UTF16_LE, BOM_UTF32_BE, BOM_UTF32_LE,
)
from os import PathLike, fspath
from os.path import realpath
from pathlib import Path
from re import DOTALL, MULTILINE, VERBOSE, Match, RegexFlag
from shutil import get_terminal_size
from typing import TYPE_CHECKING, Any, Protocol, TypeVar

from jsonyx.allow import NOTHING

if TYPE_CHECKING:
    from collections.abc import Callable, Container

    _T_co = TypeVar("_T_co", covariant=True)

    # pylint: disable-next=R0903
    class _SupportsRead(Protocol[_T_co]):
        def read(self, length: int = ..., /) -> _T_co:  # type: ignore
            """Read string."""

    _MatchFunc = Callable[[str, int], Match[str] | None]
    _Scanner = Callable[[str, str], Any]
    _StrPath = PathLike[str] | str
    _SubFunc = Callable[[str | Callable[[Match[str]], str], str], str]
    _Hook = Callable[[Any], Any]


_FLAGS: RegexFlag = VERBOSE | MULTILINE | DOTALL
_UNESCAPE: dict[str, str] = {
    '"': '"',
    "\\": "\\",
    "/": "/",
    "b": "\b",
    "f": "\f",
    "n": "\n",
    "r": "\r",
    "t": "\t",
}

_replace_unprintable: _SubFunc = re.compile(
    r"[\x00-\x1f\x7f\ud800-\udfff]", _FLAGS,
).sub
_match_chunk: _MatchFunc = re.compile(r'[^"\\\x00-\x1f]+', _FLAGS).match
_match_hex_digits: _MatchFunc = re.compile(r"[0-9A-Fa-f]{4}", _FLAGS).match
_match_line_end: _MatchFunc = re.compile(r"[^\n\r]+", _FLAGS).match
_match_number: _MatchFunc = re.compile(
    r"""
    (-?0|-?[1-9][0-9]*) # integer
    (\.[0-9]+)?         # [frac]
    ([eE][-+]?[0-9]+)?  # [exp]
    """, _FLAGS,
).match
_match_unquoted_key: _MatchFunc = re.compile(
    r"(?:\w+|[^\x00-\x7f]+)+", _FLAGS,
).match
_match_whitespace: _MatchFunc = re.compile(r"[ \t\n\r]+", _FLAGS).match


def _get_err_context(doc: str, start: int, end: int) -> tuple[int, str, int]:
    line_start: int = max(
        doc.rfind("\n", 0, start), doc.rfind("\r", 0, start),
    ) + 1
    if match := _match_whitespace(doc, line_start):
        line_start = min(match.end(), start)

    if match := _match_line_end(doc, start):
        line_end: int = match.end()
    else:
        line_end = start

    end = min(line_end, end)
    if match := _match_whitespace(doc[::-1], len(doc) - line_end):
        line_end = max(end, len(doc) - match.end())

    if end == start:
        end += 1

    max_chars: int = get_terminal_size().columns - 4  # leading spaces
    if end == line_end + 1:  # newline
        max_chars -= 1

    text_start: int = max(min(
        line_end - max_chars, end - 1 - max_chars // 2,
        start - (max_chars + 2) // 3,
    ), line_start)
    text_end: int = min(max(
        line_start + max_chars, start + (max_chars + 1) // 2,
        end + max_chars // 3,
    ), line_end)
    text: str = doc[text_start:text_end].expandtabs(1)
    text = _replace_unprintable("\ufffd", text)
    if text_start > line_start:
        text = "..." + text[3:]

    if len(text) > max_chars:
        end -= len(text) - max_chars
        text = (
            text[:max_chars // 2 - 1] + "..." + text[2 - (max_chars + 1) // 2:]
        )

    if text_end < line_end:
        text = text[:-3] + "..."

    return start - text_start + 1, text, end - text_start + 1


def _unescape_unicode(filename: str, s: str, end: int) -> int:
    if match := _match_hex_digits(s, end):
        return int(match.group(), 16)

    msg: str = "Expecting 4 hex digits"
    raise _errmsg(msg, filename, s, end, -4)


class TruncatedSyntaxError(SyntaxError):
    """Syntax error trucated to fit within the terminal width.

    .. versionadded:: 2.0

    :param msg: an error message
    :param filename: the path to the file
    :param doc: a string
    :param start: the start position
    :param end: the end position or negative offset

    Example:
        >>> import jsonyx as json
        >>> raise json.TruncatedSyntaxError(
        ...     "Invalid number", "<string>", "9" * 5000, 0, 5000
        ... )
        Traceback (most recent call last):
          File "<string>", line 1, column 1-5001
            99999999999999999999999999999999999...9999999999999999999999999999999999
            ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        jsonyx.TruncatedSyntaxError: Invalid number

    .. tip:: You can overwrite the width by setting the ``COLUMNS`` environment
        variable.

    .. seealso:: :func:`jsonyx.format_syntax_error` for formatting the
        exception.

    """

    def __init__(
        self, msg: str, filename: str, doc: str, start: int = 0, end: int = 0,
    ) -> None:
        """Create a new truncated syntax error."""
        lineno: int = (
            doc.count("\n", 0, start)
            + doc.count("\r", 0, start)
            - doc.count("\r\n", 0, start)
            + 1
        )
        colno: int = start - max(
            doc.rfind("\n", 0, start), doc.rfind("\r", 0, start),
        )
        if end <= 0:  # offset
            if match := _match_line_end(doc, start):
                end = min(match.end(), start - end)
            else:
                end = start

        end_lineno: int = (
            doc.count("\n", 0, end)
            + doc.count("\r", 0, end)
            - doc.count("\r\n", 0, end)
            + 1
        )
        end_colno: int = end - max(
            doc.rfind("\n", 0, end), doc.rfind("\r", 0, end),
        )
        offset, text, end_offset = _get_err_context(doc, start, end)
        if end_lineno != lineno:
            end_offset = end_colno

        if sys.version_info >= (3, 10):
            super().__init__(
                msg, (filename, lineno, offset, text, end_lineno, end_offset),
            )
        else:
            self.end_lineno: int = end_lineno
            self.end_offset: int = end_offset
            super().__init__(msg, (filename, lineno, offset, text))

        self.colno: int = colno
        self.end_colno: int = end_colno
        self.args = (msg, filename, doc, start, end)

    def __str__(self) -> str:
        """Convert to string."""
        if self.end_lineno == self.lineno:
            line_range: str = f"{self.lineno:d}"
        else:
            line_range = f"{self.lineno:d}-{self.end_lineno:d}"

        if self.end_colno == self.colno:
            column_range: str = f"{self.colno:d}"
        else:
            column_range = f"{self.colno:d}-{self.end_colno:d}"

        return (
            f"{self.msg} ({self.filename}, line {line_range}, column "
            f"{column_range})"
        )


TruncatedSyntaxError.__module__ = "jsonyx"


class JSONSyntaxError(TruncatedSyntaxError):
    """Invalid JSON (query) syntax."""


JSONSyntaxError.__module__ = "jsonyx"
_errmsg: type[JSONSyntaxError] = JSONSyntaxError


def detect_encoding(b: bytearray | bytes) -> str:
    r"""Detect the JSON encoding.

    :param b: a JSON string
    :return: the detected encoding

    Example:
        >>> import jsonyx as json
        >>> b = b'\x00"\x00f\x00o\x00o\x00"'
        >>> b.decode(json.detect_encoding(b))
        '"foo"'

    .. note:: Supports only ``"utf-8"``, ``"utf-8-sig"``, ``"utf-16"``,
        ``"utf-16-be"``, ``"utf-16-le"``, ``"utf-32"``, ``"utf-32-be"`` and
        ``"utf-32-le"``.
    .. seealso:: :ref:`better_decoding_error`

    """
    # JSON must start with ASCII character (not NULL)
    # Strings can't contain control characters (including NULL)
    encoding: str = "utf-8"
    startswith: Callable[[bytes | tuple[bytes, ...]], bool] = b.startswith
    if startswith((BOM_UTF32_BE, BOM_UTF32_LE)):
        encoding = "utf-32"
    elif startswith((BOM_UTF16_BE, BOM_UTF16_LE)):
        encoding = "utf-16"
    elif startswith(BOM_UTF8):
        encoding = "utf-8-sig"
    elif len(b) >= 4:
        if not b[0]:
            # 00 00 -- -- - utf-32-be
            # 00 XX -- -- - utf-16-be
            encoding = "utf-16-be" if b[1] else "utf-32-be"
        elif not b[1]:
            # XX 00 00 00 - utf-32-le
            # XX 00 00 XX - utf-16-le
            # XX 00 XX -- - utf-16-le
            encoding = "utf-16-le" if b[2] or b[3] else "utf-32-le"
    elif len(b) == 2:
        # check for cut off json
        if not b[0]:
            # 00 -- - utf-16-be
            encoding = "utf-16-be"
        elif not b[1]:
            # XX 00 - utf-16-le
            encoding = "utf-16-le"

    return encoding


detect_encoding.__module__ = "jsonyx"

try:
    if not TYPE_CHECKING:
        from _jsonyx import make_scanner
except ImportError:
    def make_scanner(
        array_hook: _Hook,
        bool_hook: _Hook,
        float_hook: _Hook,
        int_hook: _Hook,
        object_hook: _Hook,
        str_hook: _Hook,
        allow_comments: bool,
        allow_missing_commas: bool,
        allow_nan_and_infinity: bool,
        allow_surrogates: bool,
        allow_trailing_comma: bool,
        allow_unquoted_keys: bool,
        cache_keys: bool,
    ) -> _Scanner:
        """Make JSON scanner."""
        memo: dict[Any, Any] | None
        memoize: Callable[[Any, Any], Any] | None
        if cache_keys:
            memo = {}
            memoize = memo.setdefault
        else:
            memo = None
            memoize = None

        def skip_comments(filename: str, s: str, end: int) -> int:
            find: Callable[[str, int], int] = s.find
            while True:
                if match := _match_whitespace(s, end):
                    end = match.end()

                comment_idx: int = end
                if (comment_prefix := s[end:end + 2]) == "//":
                    end += 2
                    if match := _match_line_end(s, end):
                        end = match.end()
                elif comment_prefix == "/*":
                    if (end := find("*/", end + 2)) == -1:
                        if allow_comments:
                            msg: str = "Unterminated comment"
                        else:
                            msg = "Comments are not allowed"

                        raise _errmsg(msg, filename, s, comment_idx, len(s))

                    end += 2
                else:
                    return end

                if not allow_comments:
                    msg = "Comments are not allowed"
                    raise _errmsg(msg, filename, s, comment_idx, end)

        def scan_string(filename: str, s: str, end: int) -> tuple[Any, int]:
            chunks: list[str] = []
            append_chunk: Callable[[str], None] = chunks.append
            str_idx: int = end - 1
            while True:
                # Match one or more unescaped string characters
                if match := _match_chunk(s, end):
                    end = match.end()
                    append_chunk(match.group())

                # Terminator is the end of string, a literal control character,
                # or a backslash denoting that an escape sequence follows
                try:
                    terminator: str = s[end]
                except IndexError:
                    msg: str = "Unterminated string"
                    raise _errmsg(msg, filename, s, str_idx, end) from None

                if terminator == '"':
                    return str_hook("".join(chunks)), end + 1

                if terminator != "\\":
                    if terminator in {"\n", "\r"}:
                        msg = "Unterminated string"
                        raise _errmsg(msg, filename, s, str_idx, end)

                    msg = "Unescaped control character"
                    raise _errmsg(msg, filename, s, end, end + 1)

                end += 1
                try:
                    esc = s[end]
                except IndexError:
                    msg = "Expecting escaped character"
                    raise _errmsg(msg, filename, s, end) from None

                # If not a unicode escape sequence, must be in the lookup table
                if esc != "u":
                    try:
                        char = _UNESCAPE[esc]
                    except KeyError:
                        if esc in {"\n", "\r"}:
                            msg = "Expecting escaped character"
                            raise _errmsg(msg, filename, s, end) from None

                        msg = "Invalid backslash escape"
                        raise _errmsg(
                            msg, filename, s, end - 1, end + 1,
                        ) from None

                    end += 1
                else:
                    uni: int = _unescape_unicode(filename, s, end + 1)
                    end += 5
                    if 0xd800 <= uni <= 0xdbff:
                        if s[end:end + 2] == r"\u":
                            uni2: int = _unescape_unicode(filename, s, end + 2)
                            if 0xdc00 <= uni2 <= 0xdfff:
                                uni = ((uni - 0xd800) << 10) | (uni2 - 0xdc00)
                                uni += 0x10000
                                end += 6
                            elif not allow_surrogates:
                                msg = "Surrogates are not allowed"
                                raise _errmsg(msg, filename, s, end - 6, end)
                        elif not allow_surrogates:
                            msg = "Surrogates are not allowed"
                            raise _errmsg(msg, filename, s, end - 6, end)
                    elif 0xdc00 <= uni <= 0xdfff and not allow_surrogates:
                        msg = "Surrogates are not allowed"
                        raise _errmsg(msg, filename, s, end - 6, end)

                    char = chr(uni)

                append_chunk(char)

        def scan_object(filename: str, s: str, end: int) -> tuple[Any, int]:
            obj_idx: int = end - 1
            end = skip_comments(filename, s, end)
            try:
                nextchar: str = s[end]
            except IndexError:
                msg: str = "Unterminated object"
                raise _errmsg(msg, filename, s, obj_idx, end) from None

            if nextchar == "}":
                return object_hook([]), end + 1

            pairs: list[tuple[Any, Any]] = []
            append_pair: Callable[[tuple[Any, Any]], None] = pairs.append
            while True:
                key_idx: int = end
                if (nextchar := s[end:end + 1]) == '"':
                    key, end = scan_string(filename, s, end + 1)
                elif (
                    match := _match_unquoted_key(s, end)
                ) and match.group().isidentifier():
                    end = match.end()
                    if not allow_unquoted_keys:
                        msg = "Unquoted keys are not allowed"
                        raise _errmsg(msg, filename, s, key_idx, end)

                    key = match.group()
                else:
                    msg = "Expecting key"
                    raise _errmsg(msg, filename, s, end)

                # Reduce memory consumption
                if memoize is not None:
                    key = memoize(key, key)

                colon_idx: int = end
                end = skip_comments(filename, s, end)
                if s[end:end + 1] != ":":
                    msg = "Expecting colon"
                    raise _errmsg(msg, filename, s, colon_idx)

                end = skip_comments(filename, s, end + 1)
                value, end = scan_value(filename, s, end)
                append_pair((key, value))
                comma_idx: int = end
                end = skip_comments(filename, s, end)
                try:
                    nextchar = s[end]
                except IndexError:
                    msg = "Unterminated object"
                    raise _errmsg(msg, filename, s, obj_idx, end) from None

                if nextchar == ",":
                    comma_idx = end
                    end = skip_comments(filename, s, end + 1)
                elif nextchar == "}":
                    return object_hook(pairs), end + 1
                elif end == comma_idx:
                    msg = "Expecting comma"
                    raise _errmsg(msg, filename, s, comma_idx)
                elif not allow_missing_commas:
                    msg = "Missing commas are not allowed"
                    raise _errmsg(msg, filename, s, comma_idx)

                try:
                    nextchar = s[end]
                except IndexError:
                    msg = "Unterminated object"
                    raise _errmsg(msg, filename, s, obj_idx, end) from None

                if nextchar == "}":
                    if not allow_trailing_comma:
                        msg = "Trailing comma is not allowed"
                        raise _errmsg(
                            msg, filename, s, comma_idx, comma_idx + 1,
                        )

                    return object_hook(pairs), end + 1

        def scan_array(filename: str, s: str, end: int) -> tuple[Any, int]:
            arr_idx: int = end - 1
            end = skip_comments(filename, s, end)
            try:
                nextchar: str = s[end]
            except IndexError:
                msg: str = "Unterminated array"
                raise _errmsg(msg, filename, s, arr_idx, end) from None

            if nextchar == "]":
                return array_hook([]), end + 1

            values: list[Any] = []
            append_value: Callable[[Any], None] = values.append
            while True:
                value, end = scan_value(filename, s, end)
                append_value(value)
                comma_idx: int = end
                end = skip_comments(filename, s, end)
                try:
                    nextchar = s[end]
                except IndexError:
                    msg = "Unterminated array"
                    raise _errmsg(msg, filename, s, arr_idx, end) from None

                if nextchar == ",":
                    comma_idx = end
                    end = skip_comments(filename, s, end + 1)
                elif nextchar == "]":
                    return array_hook(values), end + 1
                elif end == comma_idx:
                    msg = "Expecting comma"
                    raise _errmsg(msg, filename, s, comma_idx)
                elif not allow_missing_commas:
                    msg = "Missing commas are not allowed"
                    raise _errmsg(msg, filename, s, comma_idx)

                try:
                    nextchar = s[end]
                except IndexError:
                    msg = "Unterminated array"
                    raise _errmsg(msg, filename, s, arr_idx, end) from None

                if nextchar == "]":
                    if not allow_trailing_comma:
                        msg = "Trailing comma is not allowed"
                        raise _errmsg(
                            msg, filename, s, comma_idx, comma_idx + 1,
                        )

                    return array_hook(values), end + 1

        def scan_value(filename: str, s: str, idx: int) -> tuple[Any, int]:
            try:
                nextchar = s[idx]
            except IndexError:
                msg: str = "Expecting value"
                raise _errmsg(msg, filename, s, idx) from None

            value: Any
            if nextchar == '"':
                value, end = scan_string(filename, s, idx + 1)
            elif nextchar == "{":
                try:
                    value, end = scan_object(filename, s, idx + 1)
                except RecursionError:
                    msg = "Object is too deeply nested"
                    raise _errmsg(msg, filename, s, idx) from None
            elif nextchar == "[":
                try:
                    value, end = scan_array(filename, s, idx + 1)
                except RecursionError:
                    msg = "Array is too deeply nested"
                    raise _errmsg(msg, filename, s, idx) from None
            elif nextchar == "n" and s[idx:idx + 4] == "null":
                value, end = None, idx + 4
            elif nextchar == "t" and s[idx:idx + 4] == "true":
                value, end = bool_hook(True), idx + 4  # noqa: FBT003
            elif nextchar == "f" and s[idx:idx + 5] == "false":
                value, end = bool_hook(False), idx + 5  # noqa: FBT003
            elif number := _match_number(s, idx):
                integer, frac, exp = number.groups()
                end = number.end()
                if not frac and not exp:
                    try:
                        value = int_hook(integer)
                    except Exception:  # noqa: BLE001
                        msg = "Invalid number"
                        raise _errmsg(msg, filename, s, idx, end) from None
                else:
                    try:
                        value = float_hook(
                            integer + (frac or "") + (exp or ""),
                        )
                    except Exception:  # noqa: BLE001
                        msg = "Invalid number"
                        raise _errmsg(msg, filename, s, idx, end) from None
            elif nextchar == "N" and s[idx:idx + 3] == "NaN":
                if not allow_nan_and_infinity:
                    msg = "NaN is not allowed"
                    raise _errmsg(msg, filename, s, idx, idx + 3)

                value, end = float_hook("NaN"), idx + 3
            elif nextchar == "I" and s[idx:idx + 8] == "Infinity":
                if not allow_nan_and_infinity:
                    msg = "Infinity is not allowed"
                    raise _errmsg(msg, filename, s, idx, idx + 8)

                value, end = float_hook("Infinity"), idx + 8
            elif nextchar == "-" and s[idx:idx + 9] == "-Infinity":
                if not allow_nan_and_infinity:
                    msg = "-Infinity is not allowed"
                    raise _errmsg(msg, filename, s, idx, idx + 9)

                value, end = float_hook("-Infinity"), idx + 9
            else:
                msg = "Expecting value"
                raise _errmsg(msg, filename, s, idx)

            return value, end

        def scanner(filename: str, s: str) -> Any:
            if s.startswith("\ufeff"):
                msg: str = "Unexpected UTF-8 BOM"
                raise _errmsg(msg, filename, s, 0, 1)

            end: int = skip_comments(filename, s, 0)
            try:
                obj, end = scan_value(filename, s, end)
            except JSONSyntaxError as exc:
                raise exc.with_traceback(None) from None
            finally:
                if memo is not None:
                    memo.clear()

            if (end := skip_comments(filename, s, end)) < len(s):
                msg = "Expecting end of file"
                raise _errmsg(msg, filename, s, end)

            return obj

        return scanner


class Decoder:
    """A configurable JSON decoder.

    .. versionchanged:: 2.0 Replaced ``use_decimal`` with ``hooks``.`

    .. versionchanged:: 2.2

        - Added ``cache_keys``.
        - Disabled caching keys by default.

    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param cache_keys: re-use the keys of objects
    :param hooks: the :ref:`hooks <decoding_hooks>` used for transforming data
    """

    def __init__(
        self,
        *,
        allow: Container[str] = NOTHING,
        cache_keys: bool = False,
        hooks: dict[str, _Hook] | None = None,
    ) -> None:
        """Create a new JSON decoder."""
        allow_surrogates: bool = "surrogates" in allow
        self._errors: str = "surrogatepass" if allow_surrogates else "strict"

        if hooks is None:
            hooks = {}

        self._scanner: _Scanner = make_scanner(
            hooks.get("array", list), hooks.get("bool", bool),
            hooks.get("float", float), hooks.get("int", int),
            hooks.get("object", dict), hooks.get("str", str),
            "comments" in allow, "missing_commas" in allow,
            "nan_and_infinity" in allow, allow_surrogates,
            "trailing_comma" in allow, "unquoted_keys" in allow, cache_keys,
        )

    def read(self, filename: _StrPath) -> Any:
        """Deserialize a JSON file to a Python object.

        :param filename: the path to the JSON file
        :raises OSError: if the file can't be opened
        :raises TruncatedSyntaxError: when failing to decode the file
        :return: a Python object

        Example:
            >>> import jsonyx as json
            >>> from pathlib import Path
            >>> from tempfile import TemporaryDirectory
            >>> decoder = json.Decoder()
            >>> with TemporaryDirectory() as tmpdir:
            ...     filename = Path(tmpdir) / "file.json"
            ...     _ = filename.write_text('["filesystem API"]', "utf-8")
            ...     decoder.read(filename)
            ...
            ['filesystem API']

        """
        return self.loads(Path(filename).read_bytes(), filename=filename)

    def load(
        self, fp: _SupportsRead[bytes | str], *, root: _StrPath = ".",
    ) -> Any:
        """Deserialize an open JSON file to a Python object.

        :param fp: an open JSON file
        :param root: the path to the archive containing this JSON file
        :raises TruncatedSyntaxError: when failing to decode the file
        :return: a Python object

        Example:
            >>> import jsonyx as json
            >>> from io import StringIO
            >>> decoder = json.Decoder()
            >>> io = StringIO('["streaming API"]')
            >>> decoder.load(io)
            ['streaming API']

        .. tip:: Specify ``root`` to display the zip filename in error
            messages.

        """
        name: str | None
        if name := getattr(fp, "name", None):
            return self.loads(fp.read(), filename=Path(root) / name)

        return self.loads(fp.read())

    def loads(
        self, s: bytearray | bytes | str, *, filename: _StrPath = "<string>",
    ) -> Any:
        r"""Deserialize a JSON string to a Python object.

        :param s: a JSON string
        :param filename: the path to the JSON file
        :raises TruncatedSyntaxError: when failing to decode the string
        :return: a Python object

        Example:
            >>> import jsonyx as json
            >>> decoder = json.Decoder()
            >>> decoder.loads('{"foo": ["bar", null, 1.0, 2]}')
            {'foo': ['bar', None, 1.0, 2]}
            >>> decoder.loads(r'"\"foo\bar"')
            '"foo\x08ar'

        .. tip:: Specify ``filename`` to display the filename in error
            messages.

        """
        filename = fspath(filename)
        if not filename.startswith("<") and not filename.endswith(">"):
            filename = realpath(filename)

        if not isinstance(s, str):
            encoding = detect_encoding(s)
            try:
                s = s.decode(encoding, self._errors)  # type: ignore
            except UnicodeDecodeError as exc:
                msg: str = f"(unicode error) {exc}"
                doc: str = exc.object.decode(encoding, "replace")
                # pylint: disable-next=E1127
                start: str = exc.object[:exc.start].decode(encoding, "replace")
                # pylint: disable-next=E1127
                end: str = exc.object[:exc.end].decode(encoding, "replace")
                raise TruncatedSyntaxError(
                    msg, filename, doc, len(start), len(end),
                ) from None

        return self._scanner(filename, s)  # type: ignore


Decoder.__module__ = "jsonyx"



"""JSON encoder."""
from __future__ import annotations

__all__: list[str] = ["Encoder"]

import re
import sys
from io import StringIO
from pathlib import Path
from re import DOTALL, MULTILINE, VERBOSE, Match, RegexFlag
from typing import TYPE_CHECKING, Any, Protocol, TypeVar

from jsonyx import TruncatedSyntaxError
from jsonyx.allow import NOTHING

if TYPE_CHECKING:
    from collections.abc import Callable, Container, ItemsView
    from os import PathLike

    _T = TypeVar("_T")
    _T_contra = TypeVar("_T_contra", contravariant=True)

    # pylint: disable-next=R0903
    class _SupportsWrite(Protocol[_T_contra]):
        def write(self, s: _T_contra, /) -> object:
            """Write string."""

    _EncodeFunc = Callable[[_T], str]
    _Hook = Callable[[Any], Any]
    _MatchFunc = Callable[[str], Match[str] | None]
    _StrPath = PathLike[str] | str
    _SubFunc = Callable[[str | Callable[[Match[str]], str], str], str]
    _WriteFunc = Callable[[str], object]


_ESCAPE_DCT: dict[str, str] = {
    **{chr(i): f"\\u{i:04x}" for i in range(0x20)},
    '"': '\\"',
    "\\": "\\\\",
    "\b": "\\b",
    "\f": "\\f",
    "\n": "\\n",
    "\r": "\\r",
    "\t": "\\t",
}
_FLAGS: RegexFlag = VERBOSE | MULTILINE | DOTALL

_escape: _SubFunc = re.compile(r'["\\\x00-\x1f]', _FLAGS).sub
_escape_ascii: _SubFunc = re.compile(r'["\\]|[^\x20-\x7e]', _FLAGS).sub
_match_number: _MatchFunc = re.compile(
    r"""
    (-?0|-?[1-9][0-9]*) # integer
    (\.[0-9]+)?         # [frac]
    ([eE][-+]?[0-9]+)?  # [exp]
    """, _FLAGS,
).fullmatch

try:
    if not TYPE_CHECKING:
        from _jsonyx import make_encoder
except ImportError:
    def make_encoder(
        array_types: type | tuple[type, ...],
        bool_types: type | tuple[type, ...],
        float_types: type | tuple[type, ...],
        hook: _Hook | None,
        indent: str | None,
        int_types: type | tuple[type, ...],
        object_types: type | tuple[type, ...],
        str_types: type | tuple[type, ...],
        end: str,
        item_separator: str,
        key_separator: str,
        long_item_separator: str,
        max_indent_level: int,
        allow_nan_and_infinity: bool,
        allow_non_str_keys: bool,
        allow_surrogates: bool,
        check_circular: bool,
        ensure_ascii: bool,
        indent_leaves: bool,
        quoted_keys: bool,
        skipkeys: bool,
        sort_keys: bool,
        trailing_comma: bool,
    ) -> _EncodeFunc[object]:
        """Make JSON encoder."""
        markers: dict[int, object] | None = {} if check_circular else None

        if hook is None:
            def new_hook(obj: Any) -> Any:
                return obj

            hook = new_hook

        if not ensure_ascii:
            def replace(match: Match[str]) -> str:
                return _ESCAPE_DCT[match.group()]

            def encode_string(s: str) -> str:
                return f'"{_escape(replace, s)}"'
        else:
            def replace(match: Match[str]) -> str:
                s: str = match.group()
                try:
                    return _ESCAPE_DCT[s]
                except KeyError:
                    if (uni := ord(s)) >= 0x10000:
                        # surrogate pair
                        uni -= 0x10000
                        uni1: int = 0xd800 | ((uni >> 10) & 0x3ff)
                        uni2: int = 0xdc00 | (uni & 0x3ff)
                        return f"\\u{uni1:04x}\\u{uni2:04x}"

                    if 0xd800 <= uni <= 0xdfff and not allow_surrogates:
                        msg: str = "Surrogates are not allowed"
                        raise ValueError(msg) from None

                    return f"\\u{uni:04x}"

            def encode_string(s: str) -> str:
                return f'"{_escape_ascii(replace, s)}"'

        def encode_float(num: Any) -> str:
            s: str = str(num)
            if _match_number(s):
                return s

            if s.lower() == "nan":
                s = "NaN"
            elif s.lower() in {"inf", "infinity"}:
                s = "Infinity"
            elif s.lower() in {"-inf", "-infinity"}:
                s = "-Infinity"
            else:
                msg: str = f"{num!r} is not JSON serializable"
                raise ValueError(msg)

            if not allow_nan_and_infinity:
                msg = f"{num!r} is not allowed"
                raise ValueError(msg)

            return s

        def write_sequence(
            seq: Any, write: _WriteFunc, indent_level: int, old_indent: str,
        ) -> None:
            if markers is not None:
                if (markerid := id(seq)) in markers:
                    msg: str = "Unexpected circular reference"
                    raise ValueError(msg)

                markers[markerid] = seq

            write("[")
            current_indent: str = old_indent
            if indent is None or indent_level >= max_indent_level or (
                not indent_leaves and not any(isinstance(hook(value), (
                    list, tuple, dict, array_types, object_types,
                )) for value in seq)
            ):
                indented: bool = False
                current_item_separator: str = long_item_separator
            else:
                indented = True
                indent_level += 1
                current_indent += indent
                current_item_separator = item_separator + current_indent

            first: bool = True
            for value in seq:
                if first:
                    first = False
                    if indented:
                        write(current_indent)
                else:
                    write(current_item_separator)

                write_value(value, write, indent_level, current_indent)

            if markers is not None:
                del markers[markerid]  # type: ignore

            if not first and indented:
                if trailing_comma:
                    write(item_separator)

                write(old_indent)

            write("]")

        def write_mapping(
            mapping: Any,
            write: _WriteFunc,
            indent_level: int,
            old_indent: str,
        ) -> None:
            if markers is not None:
                if (markerid := id(mapping)) in markers:
                    msg: str = "Unexpected circular reference"
                    raise ValueError(msg)

                markers[markerid] = mapping

            write("{")
            current_indent: str = old_indent
            if indent is None or indent_level >= max_indent_level or (
                not indent_leaves and not any(isinstance(hook(value), (
                    list, tuple, dict, array_types, object_types,
                )) for value in mapping.values())
            ):
                indented: bool = False
                current_item_separator: str = long_item_separator
            else:
                indented = True
                indent_level += 1
                current_indent += indent
                current_item_separator = item_separator + current_indent

            first: bool = True
            items: ItemsView[object, object] = mapping.items()
            for key, value in sorted(items) if sort_keys else items:
                key = hook(key)  # noqa: PLW2901
                if isinstance(key, (str, str_types)):
                    s = str(key)
                else:
                    if key is None:
                        s = "null"
                    elif isinstance(key, (bool, bool_types)):
                        s = "true" if key else "false"
                    elif isinstance(key, (float, int, float_types, int_types)):
                        s = encode_float(key)
                    elif skipkeys:
                        continue
                    else:
                        msg = f"Keys must be str, not {type(key).__name__}"
                        raise TypeError(msg)

                    if not allow_non_str_keys:
                        if skipkeys:
                            continue

                        msg = "Non-string keys are not allowed"
                        raise TypeError(msg)

                if first:
                    first = False
                    if indented:
                        write(current_indent)
                else:
                    write(current_item_separator)

                if not quoted_keys and s.isidentifier() and (
                    not ensure_ascii or s.isascii()
                ):
                    write(s)
                else:
                    write(encode_string(s))

                write(key_separator)
                write_value(value, write, indent_level, current_indent)

            if markers is not None:
                del markers[markerid]  # type: ignore

            if not first and indented:
                if trailing_comma:
                    write(item_separator)

                write(old_indent)

            write("}")

        def write_value(
            obj: object,
            write: _WriteFunc,
            indent_level: int,
            current_indent: str,
        ) -> None:
            obj = hook(obj)
            if obj is None:
                write("null")
            elif isinstance(obj, (bool, bool_types)):
                write("true" if obj else "false")
            elif isinstance(obj, (str, str_types)):
                write(encode_string(str(obj)))
            elif isinstance(obj, (float, int, float_types, int_types)):
                write(encode_float(obj))
            elif isinstance(obj, (list, tuple, array_types)):
                write_sequence(obj, write, indent_level, current_indent)
            elif isinstance(obj, (dict, object_types)):
                write_mapping(obj, write, indent_level, current_indent)
            else:
                msg: str = f"{type(obj).__name__} is not JSON serializable"
                raise TypeError(msg)

        def encoder(obj: object) -> str:
            io: StringIO = StringIO()
            write: _WriteFunc = io.write
            try:
                write_value(obj, write, 0, "\n")
            except (ValueError, TypeError) as exc:
                raise exc.with_traceback(None) from None
            finally:
                if markers is not None:
                    markers.clear()

            write(end)
            return io.getvalue()

        return encoder


class Encoder:
    r"""A configurable JSON encoder.

    .. versionchanged:: 2.0

        - Added ``commas``, ``indent_leaves``, ``max_indent_level``,
          ``quoted_keys`` and ``types``.
        - Made :class:`tuple` serializable by default instead of
          :class:`enum.Enum` and :class:`decimal.Decimal`.
        - Replaced ``item_separator`` and ``key_separator`` with
          ``separators``.

    .. versionchanged:: 2.1 Added ``check_circular``, ``hook`` and
        ``skipkeys``.

    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param check_circular: check for circular references
    :param commas: separate items by commas when indented
    :param hook: the :ref:`hook <encoding_hook>` used for transforming data
    :param end: the string to append at the end
    :param ensure_ascii: escape non-ASCII characters
    :param indent: the number of spaces or string to indent with
    :param indent_leaves: indent leaf objects and arrays
    :param max_indent_level: the level up to which to indent
    :param quoted_keys: quote keys which are :ref:`identifiers <identifiers>`
    :param separators: the item and key separator
    :param skipkeys: skip non-string keys
    :param sort_keys: sort the keys of objects
    :param trailing_comma: add a trailing comma when indented
    :param types: a dictionary of :ref:`additional types <protocol_types>`

    .. note:: The item separator is automatically stripped when indented.
    .. warning:: Avoid specifying ABCs for ``types``, that is very slow.
    """

    def __init__(
        self,
        *,
        allow: Container[str] = NOTHING,
        check_circular: bool = True,
        commas: bool = True,
        hook: _Hook | None = None,
        end: str = "\n",
        ensure_ascii: bool = False,
        indent: int | str | None = None,
        indent_leaves: bool = True,
        max_indent_level: int | None = None,
        quoted_keys: bool = True,
        separators: tuple[str, str] = (", ", ": "),
        skipkeys: bool = False,
        sort_keys: bool = False,
        trailing_comma: bool = False,
        types: dict[str, type | tuple[type, ...]] | None = None,
    ) -> None:
        """Create a new JSON encoder."""
        allow_surrogates: bool = "surrogates" in allow
        long_item_separator, key_separator = separators
        if commas:
            item_separator: str = long_item_separator.rstrip()
        else:
            item_separator = ""

        if indent is not None and isinstance(indent, int):
            indent *= " "

        if max_indent_level is None:
            max_indent_level = sys.maxsize

        if types is None:
            types = {}

        self._encoder: _EncodeFunc[object] = make_encoder(
            types.get("array", ()), types.get("bool", ()),
            types.get("float", ()), hook, indent, types.get("int", ()),
            types.get("object", ()), types.get("str", ()), end,
            item_separator, key_separator, long_item_separator,
            max_indent_level, "nan_and_infinity" in allow,
            "non_str_keys" in allow, allow_surrogates, check_circular,
            ensure_ascii, indent_leaves, quoted_keys, skipkeys, sort_keys,
            commas and trailing_comma,
        )
        self._errors: str = "surrogatepass" if allow_surrogates else "strict"

    def write(
        self, obj: object, filename: _StrPath, encoding: str = "utf-8",
    ) -> None:
        r"""Serialize a Python object to a JSON file.

        .. versionchanged:: 2.0 Added ``encoding``.

        :param obj: a Python object
        :param filename: the path to the JSON file
        :param encoding: the JSON encoding
        :raises OSError: if the file can't be opened
        :raises RecursionError: if the object is too deeply nested
        :raises TypeError: for unserializable values
        :raises TruncatedSyntaxError: when failing to encode the file
        :raises ValueError: for invalid values

        Example:
            >>> import jsonyx as json
            >>> from pathlib import Path
            >>> from tempfile import TemporaryDirectory
            >>> encoder = json.Encoder()
            >>> with TemporaryDirectory() as tmpdir:
            ...     filename = Path(tmpdir) / "file.json"
            ...     encoder.write(["filesystem API"], filename)
            ...     filename.read_text("utf-8")
            ...
            '["filesystem API"]\n'

        """
        with Path(filename).open("w", -1, encoding, self._errors) as fp:
            self.dump(obj, fp)

    def dump(self, obj: object, fp: _SupportsWrite[str] | None = None) -> None:
        r"""Serialize a Python object to an open JSON file.

        :param obj: a Python object
        :param fp: an open JSON file
        :raises RecursionError: if the object is too deeply nested
        :raises TypeError: for unserializable values
        :raises TruncatedSyntaxError: when failing to write to the file
        :raises ValueError: for invalid values

        Example:
            Writing to standard output:

            >>> import jsonyx as json
            >>> encoder = json.Encoder()
            >>> import jsonyx as json
            >>> encoder.dump('"foo\bar')
            "\"foo\bar"
            >>> encoder.dump("\\")
            "\\"
            >>> encoder.dump("\u20AC")
            "€"

            Writing to an open file:

            >>> import jsonyx as json
            >>> from io import StringIO
            >>> io = StringIO()
            >>> encoder.dump(["streaming API"], io)
            >>> io.getvalue()
            '["streaming API"]\n'

        """
        s: str = self._encoder(obj)
        try:
            if fp is None:
                sys.stdout.write(s)  # Use sys.stdout to work with doctest
            else:
                fp.write(s)
        except UnicodeEncodeError as exc:
            msg: str = f"(unicode error) {exc}"
            raise TruncatedSyntaxError(
                msg, "<string>", exc.object, exc.start, exc.end,
            ) from None

    def dumps(self, obj: object) -> str:
        r"""Serialize a Python object to a JSON string.

        :param obj: a Python object
        :raises RecursionError: if the object is too deeply nested
        :raises TypeError: for unserializable values
        :raises ValueError: for invalid values
        :return: a JSON string

        Example:
            >>> import jsonyx as json
            >>> encoder = json.Encoder()
            >>> encoder.dumps(["foo", {"bar": ("baz", None, 1.0, 2)}])
            '["foo", {"bar": ["baz", null, 1.0, 2]}]\n'

        """
        return self._encoder(obj)


Encoder.__module__ = "jsonyx"
